/* $Id: omega_coadd.c,v 1.6 2011-06-27 12:59:41 agabasch Exp $
 *
 * This file is part of the OMEGA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: agabasch $
 * $Date: 2011-06-27 12:59:41 $
 * $Revision: 1.6 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
   								Includes
 -----------------------------------------------------------------------------*/
#include <string.h>
#include <math.h>
#include <stdio.h>


#include "omega_coadd.h"
#include "omega_fits.h"
#include "omega_dfs.h"
#include "omega_utils.h"
#include "omega_pfits.h"

/**
 * @defgroup omega_coadd  Stacking Related Functions
 *
 * \par Synopsis
 * Function to stack a series of images, and to get the offset keywords from
 * observations.
 *
 * \code
 * #include omega_coadd.h
 * #include omega_fits.h
 * #include omega_dfs.h
 * #include omega_utils.h
 * #include omega_pfits.h
 * \endcode
 */

/**@{*/
/**
 * @brief		Create a stack image and a confidence map from a series
 *              of jittered images
 *
 * @param	jset	set of frames with jittered frames only
 * @param	offsets	array of bivectors with offsets for all images
 * @params	pars	parameter list
 * @params	ext		FITS extension to load
 *
 * @return	array with two images, the first is the stack the second is
 * the confidence map. Returns NULL in case of error.
 */
cpl_image **omega_jitter_stack(cpl_frameset *jset, cpl_frameset *jset_mask, cpl_bivector *offsets, int refine,
						cpl_parameterlist *pars, int ext)
{

	int n = 0;
	int n_mask=0;
	int i=0;
	cpl_mask * mask=NULL;
    double psigmas[] = {10.0, 5.0, 2.0};
    const int nsigmas = (int)(sizeof(psigmas)/sizeof(double));
    int shx, shy, mhx, mhy, minrej, maxrej;
    const char *sval;
    cpl_geom_combine comb_meth;
	cpl_imagelist *ilist;
	cpl_imagelist *ilist_mask;
	cpl_vector *sigmas;
	cpl_parameter *par;
	cpl_image **combined = NULL;
    cpl_vector * offset_x = NULL;
    cpl_vector * offset_y = NULL;

	if((jset == NULL) || (pars == NULL))
		return NULL;

	/* Get parameters */
	par = cpl_parameterlist_find(pars, "omega.omega_science.xcorr") ;
	sval = cpl_parameter_get_string(par);

    if (sscanf(sval, "%d,%d,%d,%d",&shx,&shy,&mhx,&mhy)!=4)
        return NULL;

    par = cpl_parameterlist_find(pars, "omega.omega_science.CombMode");
    sval = cpl_parameter_get_string(par);
    if (!strcmp(sval, "union"))
    	comb_meth = CPL_GEOM_UNION;
    else if (!strcmp(sval, "intersect"))
    	comb_meth = CPL_GEOM_INTERSECT;
    else if (!strcmp(sval, "first"))
    	comb_meth = CPL_GEOM_FIRST;
    else
    	return NULL;

    /* Overule combination method if offsets are too big */

    offset_x=cpl_bivector_get_x(offsets);
    offset_y=cpl_bivector_get_y(offsets);

    if (cpl_vector_get_max(offset_x)-cpl_vector_get_min(offset_x) > 3000 ||
    		cpl_vector_get_max(offset_y)-cpl_vector_get_min(offset_y) > 6000)
    {
    	cpl_msg_warning(cpl_func, "Dithering offsets are very large. "
    			"Combination method changed to \"first\" ");
    	comb_meth = CPL_GEOM_FIRST;
    }

    par = cpl_parameterlist_find(pars, "omega.omega_science.StackRej");
    sval = cpl_parameter_get_string(par);
    if(sscanf(sval, "%d,%d",&minrej,&maxrej) !=2)
    	return NULL;

    n = cpl_frameset_get_size(jset);
    n_mask = cpl_frameset_get_size(jset_mask);

	if(n != n_mask){
		cpl_vector_unwrap(sigmas);
		cpl_msg_warning(cpl_func,"Number of images and masks differ!");
		return NULL;
	}


	/* Create vector of sigmas for aperture extraction */
	sigmas = cpl_vector_wrap(nsigmas, psigmas);

	/* Load frameset into an image list */
	ilist = cpl_imagelist_load_frameset(jset, CPL_TYPE_FLOAT, 1, ext);
	if(ilist == NULL){
		cpl_vector_unwrap(sigmas);
		return NULL;
	}

	ilist_mask = cpl_imagelist_load_frameset(jset_mask, CPL_TYPE_INT, 1, ext);
	if(ilist_mask == NULL){
		cpl_vector_unwrap(sigmas);
		return NULL;
	}

	/*Add mask files to the images*/

	for (i=0; i<n; i++){
	mask = cpl_mask_threshold_image_create(cpl_imagelist_get(ilist_mask,i), 0.5, 1.5);
	cpl_image_reject_from_mask(cpl_imagelist_get(ilist,i), mask);
	freemask(mask);
	}

	freeilist(ilist_mask);
	/*
	 * Create stack and confidence map
	 */
	combined = cpl_geom_img_offset_combine(ilist,offsets,refine,NULL,sigmas,NULL,
							shx,shy,mhx,mhy,minrej,maxrej,comb_meth);

	if(combined == NULL){
		cpl_msg_debug(cpl_func,"Error in combination function. %s",cpl_error_get_message());
		cpl_vector_unwrap(sigmas);
		freeilist(ilist);
		return NULL;
	}

	/* Clean up */
	cpl_vector_unwrap(sigmas);
	freeilist(ilist);

	return combined;
}

/**
 * @brief		Get keyword values of stare/jitter/dither offsets
 *
 * @param	set		Frame set with images
 * @param   mode    Science mode (1=stare,2=jitter,3=dither,4=offset)
 *
 * @return	offsets	bivector with offsets for all images of
 * 					the frame set. NULL in case of error.
 *
 * Bivector is created as X,Y, one pair per image.
 */
cpl_bivector *omega_tel_offsets(cpl_frameset *set, int mode)
{

	int i = 0;
	double *offsets_x, *offsets_y;
	double  scale = 0.0;
	cpl_frame *frame;
	cpl_propertylist *plist;
	cpl_bivector *offsets;

	if((set == NULL) || (mode <= 0) || (mode > 4))
		return NULL;

	/* Get pixel scale of image, in arcsec */
	frame = cpl_frameset_get_first(set);
	plist = cpl_propertylist_load(cpl_frame_get_filename(frame), 1);

	if((scale = omega_pfits_get_cd22(plist)) != 0)
		scale = scale*3600.;
	else if((scale = omega_pfits_get_cdelt2(plist)) != 0)
		scale = scale*3600;

	freeplist(plist);

	/* Allocate space for offsets */
	offsets = cpl_bivector_new(cpl_frameset_get_size(set));
	offsets_x = cpl_bivector_get_x_data(offsets);
	offsets_y = cpl_bivector_get_y_data(offsets);

	/* Set reference offsets to 0 */
	/* Here we are assuming that first image is the reference
	 * with X=Y=0. Is this correct?
	 */
    offsets_x[0] = 0;
    offsets_y[0] = 0;
    frame = cpl_frameset_get_next(set);

    /* Starts from second image */
    /* Write offsets in pixels */
    for(i=1; i<cpl_frameset_get_size(set); i++){
    	plist = cpl_propertylist_load(cpl_frame_get_filename(frame), 0);
    	if(mode == 1){
    		/* STARE mode */
    		offsets_x[i] = offsets_y[i] = 0.0;
    	}
    	else if(mode == 2){
    		offsets_x[i] = omega_pfits_get_jitter_offset(plist)/scale;
    		/* For jitter mode, offsets X=Y */
    		offsets_y[i] = offsets_x[i];
    	}
    	else if(mode == 3){
    		/* DITHER mode */
    		offsets_x[i] = omega_pfits_get_dither_offsetx(plist)/scale;
    		offsets_y[i] = omega_pfits_get_dither_offsety(plist)/scale;
    	}
    	else if(mode == 4){
    		/* OFFSET mode */
    		offsets_x[i] = omega_pfits_get_offset_alpha(plist)/scale;
    		offsets_y[i] = omega_pfits_get_offset_delta(plist)/scale;
    	}
    	frame = cpl_frameset_get_next(set);
    	freeplist(plist);
    	cpl_msg_debug(cpl_func,"im=%d offset_x=%g offset_y=%g",i,offsets_x[i],offsets_y[i]);
    }

    /* TEST */
//    offsets_x[1] = -130;
//    offsets_y[1] = 100;
//    offsets_x[2] = 134;
//    offsets_y[2] = 197;



	return offsets;
}

/*----------------------------------------------------------------------------*/
/**
   @brief    Get the offsets of a set of images with respect to mask.
   @param    frameset   frameset of jittered images.
   @param    ref_frame  reference image frame.
   @param    ext        extension number, 0 for primary header.
   @return   Telescope offsets in a bivector, of NULL in case of error.

   Note that the telescope offsets are already multiplied with -1 to get
   to the CPL convention. Function cloned from mime_get_offsets_from_wcs
**/
/*----------------------------------------------------------------------------*/
cpl_bivector *omega_get_offsets_from_wcs (
    const cpl_frameset * frameset,
    const cpl_frame * ref_frame,
    int ext)
{
    cpl_bivector    * offsets;
    double          * offsets_x;
    double          * offsets_y;
    const cpl_frame * this_frame;
    int               nfiles;
    int               i;
    cpl_propertylist *plist;
    cpl_wcs  *ref_wcs;

    /* Test entries */
    if (frameset == NULL){
        cpl_error_set(cpl_func, CPL_ERROR_NULL_INPUT);
        return NULL;
    }

    /* Create the offsets bivector */
    nfiles = cpl_frameset_get_size(frameset);
    offsets = cpl_bivector_new(nfiles);
    offsets_x = cpl_bivector_get_x_data(offsets);
    offsets_y = cpl_bivector_get_y_data(offsets);

    /* WCS of reference frame */
    /* The nonlinear PV parameters cause WCSLIB to hiccup, so we
       omit them from the propertylist */
    plist = cpl_propertylist_load_regexp(
        cpl_frame_get_filename(ref_frame),
        ext, "PV", 1);
    ref_wcs = cpl_wcs_new_from_propertylist(plist);
    /* Get the WCS info */
    if(cpl_error_get_code() == CPL_ERROR_NO_WCS)
    {
        cpl_msg_error(__func__,"Not compiled with WCS support.");
        cpl_propertylist_delete(plist);
        return NULL;
    }


    cpl_propertylist_delete(plist);

    /* Check whether WCS could be loaded */
    if (ref_wcs == NULL){
        cpl_msg_error(cpl_func, "Could not read WCS from %s",
                      cpl_frame_get_filename(ref_frame));
        cpl_bivector_delete(offsets);
        return NULL;
    }

    for (i = 0; i < nfiles; i++){
        cpl_matrix       * exp_pixel;
        cpl_matrix       * exp_equat;
        cpl_matrix       * ref_pixel;
        cpl_array        * status =NULL;
        cpl_wcs          * exp_wcs;
        cpl_error_code     error = CPL_ERROR_NONE;

        /* Pixel (1,1) */
        exp_pixel = cpl_matrix_new(1, 2);
        cpl_matrix_set(exp_pixel, 0, 0, 1.);
        cpl_matrix_set(exp_pixel, 0, 1, 1.);

        /* WCS of exposure i */
        this_frame = cpl_frameset_get_frame_const(frameset, i);
        plist = cpl_propertylist_load_regexp(
            cpl_frame_get_filename(this_frame),
            ext, "PV", 1);
        exp_wcs = cpl_wcs_new_from_propertylist(plist);
        cpl_propertylist_delete(plist);

        /* Check whether WCS could be loaded */
        if (exp_wcs == NULL){
            cpl_msg_error(cpl_func, "Could not read WCS from %s",
                          cpl_frame_get_filename(this_frame));
            cpl_matrix_delete(exp_pixel);
            cpl_bivector_delete(offsets);
            return NULL;
        }

        /* transform pixel (1,1) to corresponding pixel in
           reference exposure */
        status = NULL;
        if ((error = cpl_wcs_convert(exp_wcs, exp_pixel, &exp_equat, &status,
                                     CPL_WCS_PHYS2WORLD))
            != CPL_ERROR_NONE){
            cpl_msg_error(cpl_func, cpl_error_get_message_default(error));
            cpl_matrix_delete(exp_pixel);
            cpl_wcs_delete(exp_wcs);
            cpl_bivector_delete(offsets);
            cpl_array_delete(status);
            return NULL;
        }
        cpl_matrix_delete(exp_pixel);
        cpl_array_delete(status);

        status = NULL;
        if ((error = cpl_wcs_convert(ref_wcs, exp_equat, &ref_pixel, &status,
                                     CPL_WCS_WORLD2PHYS))
            != CPL_ERROR_NONE){
            cpl_msg_error(cpl_func, cpl_error_get_message_default(error));
            cpl_matrix_delete(exp_equat);
            cpl_wcs_delete(exp_wcs);
            cpl_bivector_delete(offsets);
            cpl_array_delete(status);
            return NULL;
        }
        cpl_matrix_delete(exp_equat);
        cpl_wcs_delete(exp_wcs);
        cpl_array_delete(status);

        /* get offsets */
        offsets_x[i] = cpl_matrix_get(ref_pixel, 0, 0);
        offsets_y[i] = cpl_matrix_get(ref_pixel, 0, 1);

        cpl_matrix_delete(ref_pixel);
    }

    /* On success */
    cpl_wcs_delete(ref_wcs);
    return offsets;
}


/**@}*/
