/* $Id: omega_fftw.c,v 1.3 2012-01-31 13:38:50 agabasch Exp $
 *
 * This file is part of the OMEGA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: agabasch $
 * $Date: 2012-01-31 13:38:50 $
 * $Revision: 1.3 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include <complex.h>

#include "omega_bpm.h"

#include "omega_utils.h"
/**
 * @defgroup omega_fftw  Fast Fourier Transformation
 *
 * This module provides functions to do fast fourier transformations
 *
 * @par Synopsis:
 *   Functions to deal with fast fourier transformations
 *
 * @code
 * #include "omega_bpm.h"
 * #include "omega_utils.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/


/*-----------------------------------------------------------------------------
                                Defines
 -----------------------------------------------------------------------------*/
cpl_image *omega_mirror_edges(cpl_image *image, int dx, int dy);

/*-------------------------------------------------------------------------*/
/**
@brief	Generate a low pass filter for FFT convolution .
@param	xs	x size of the generated image.
@param	ys	y size of the generated image.
@param	sigma_x	Sigma for the gaussian distribution.
@param	sigma_y      Sigma for the gaussian distribution.
@return	1 newly allocated image.

This function generates an image of a 2d gaussian, modified in such
a way that the different quadrants have a quadrants of the gaussian
in the corner. This image is suitable for FFT convolution.
Copied from eclipse, src/iproc/generate.c

The returned image must be deallocated.
*/
/*--------------------------------------------------------------------------*/
cpl_image * omega_gen_lowpass(int xs, int ys, double sigma_x, double sigma_y)
{

	int i= 0.0;
	int j= 0.0;
	int hlx= 0.0;
	int hly = 0.0;
	double x= 0.0;
	double y= 0.0;
	double gaussval= 0.0;
	float *data;

	cpl_image 	*lowpass_image;


	lowpass_image = cpl_image_new (xs, ys, CPL_TYPE_FLOAT);
	if (lowpass_image == NULL) {
		cpl_msg_error (cpl_func, "Cannot generate lowpass filter <%s>",cpl_error_get_message());
		return NULL;
	}

	hlx = xs/2;
	hly = ys/2;

	data = cpl_image_get_data_float(lowpass_image);

	/* Given an image with pixels 0<=i<N, 0<=j<M then the convolution image
   has the following properties:

   ima[0][0] = 1
   ima[i][0] = ima[N-i][0] = exp (-0.5 * (i/sig_i)^2)   1<=i<N/2
   ima[0][j] = ima[0][M-j] = exp (-0.5 * (j/sig_j)^2)   1<=j<M/2
   ima[i][j] = ima[N-i][j] = ima[i][M-j] = ima[N-i][M-j]
             = exp (-0.5 * ((i/sig_i)^2 + (j/sig_j)^2))
	 */

	data[0] = (float)1.0;

	/* first row */
	for (i=1 ; i<=hlx ; i++) {
		x = (double)i / sigma_x;
		gaussval = (double)exp(-0.5*x*x);
		data[i] = gaussval;
		data[xs-i] = gaussval;
	}

	for (j=1; j<=hly ; j++) {
		y = (double)j / sigma_y;
		/* first column */
		data[j*xs] = (double)exp(-0.5*y*y);
		data[(ys-j)*xs] = (double)exp(-0.5*y*y);

		for (i=1 ; i<=hlx ; i++) {
			/* Use internal symetries */
			x = (double) i / sigma_x;
			gaussval = (double)exp (-0.5*(x*x+y*y));
			data[j*xs+i] = gaussval;
			data[(j+1)*xs-i] = gaussval;
			data[(ys-j)*xs+i] = gaussval;
			data[(ys+1-j)*xs-i] = gaussval;

		}
	}

	/* FIXME: for the moment, reset errno which is coming from exp()
            in first for-loop at i=348. This is causing cfitsio to
            fail when loading an extension image (bug in cfitsio too).
	 */
	if(errno != 0)
		errno = 0;

	return lowpass_image;
}

/**      Function to calculate the low spatial frequency
*
* @brief        Get low spatial frequency componenets from the flat field
* 				 using the FFTW
* @param   flat       twilight or dome flat
* @param   bpm        optonal bad pixels image
* @param	gausfilt   Gaussian Fourier filter size
* @param	mirrorx	   for mirroring edges (ocfft continuity)
* @param	mirrory	   for mirroring edges (ocfft continuity)
* @param	medfilt for cleaning bad pixels
*
* @return       1 newly allocated image.
*/
cpl_image * omega_get_spatial_freq(cpl_image *flat, cpl_mask *bpm,
		double gausfilt, int mirrorx, int mirrory){

	int xsize=0, ysize=0;

	double sigma_x = 0.;
	double sigma_y = 0.;

	cpl_image *clean_flat=NULL;
	cpl_image *eflat=NULL;
	cpl_image *eflat_complex=NULL;
	cpl_image *eflat_real=NULL;

	cpl_image *filter_image=NULL;
	cpl_image *filter_image_complex=NULL;
	cpl_image *flat_real;


	/*Clean flat from bad pixels if bpm exists*/
	clean_flat = cpl_image_duplicate(flat);
	if (bpm != NULL) {
		cpl_image_reject_from_mask(clean_flat,bpm);
		cpl_detector_interpolate_rejected(clean_flat);
	}

	/* Expand the flat image using the mirror edges function */
	eflat = omega_mirror_edges(clean_flat, mirrorx, mirrory);
	if(eflat == NULL){
		freeimage(clean_flat);
		cpl_msg_error(cpl_func,"Filter image is NULL");
		return NULL;
	}

	freeimage(clean_flat);

	xsize = cpl_image_get_size_x(eflat);
	ysize = cpl_image_get_size_y(eflat);

	sigma_x = gausfilt;
	sigma_y = (double)(sigma_x * ysize) / xsize;


	/* Generate a lowpass filter to be used in the FFT convolution */
	filter_image = omega_gen_lowpass(xsize, ysize, sigma_x, sigma_y);
	if(filter_image == NULL){
		cpl_msg_error(cpl_func,"Filter image is NULL");
		return NULL;
	}

	eflat_complex = cpl_image_new(xsize,ysize, CPL_TYPE_FLOAT_COMPLEX);
	eflat_real = cpl_image_new(xsize,ysize, CPL_TYPE_FLOAT);
	filter_image_complex =cpl_image_cast(filter_image,CPL_TYPE_FLOAT_COMPLEX);

	/*Free memory*/
	cpl_image_delete(filter_image);


	/* Apply a forward FFT on the images  */
	cpl_fft_image(eflat_complex, eflat, CPL_FFT_FORWARD);
	/*Free memory*/
	cpl_image_delete(eflat);

	/*Multiply the filter with the the FFT image */
	cpl_image_multiply(eflat_complex,filter_image_complex);


	/* Apply a backward FFT on the images  */
	cpl_fft_image(eflat_real, eflat_complex,CPL_FFT_BACKWARD);
	/*Free memory*/
	cpl_image_delete(eflat_complex);
	cpl_image_delete(filter_image_complex);

	/*  Extract original image from the expanded image.  */
	flat_real = cpl_image_extract(eflat_real, mirrorx+1, mirrory+1,
			xsize-mirrorx, ysize-mirrory);

	if (flat_real == NULL) {
		cpl_msg_error (cpl_func,"Real extracted image is NULL. <%s>", cpl_error_get_message());
		return NULL;
	}
	cpl_image_delete(eflat_real);


	return flat_real;
}


/**
@brief    expand image by mirroring edges
@param    image_in      Image.
@param    x_size        the number of pixels in x to expand
@param    y_size        the number of pixels in y to expand

*/
/*--------------------------------------------------------------------------*/
cpl_image *omega_mirror_edges(cpl_image *image, int dx, int dy)
{

	int xx = 0;
	int yy = 0;
	int xs = 0;
	int ys = 0;
	int inrow = 1;
	int outrow = 1;
	int i =0;
	int j = 0;

	float *data;
	float *out_data;


	cpl_image 	*big_image;



	xs = cpl_image_get_size_x(image);
	ys = cpl_image_get_size_y(image);

	xx = xs+(2*dx);
	yy = ys+(2*dy);

	data = cpl_image_get_data_float(image);

	big_image = cpl_image_new(xx, yy, CPL_TYPE_FLOAT);
	out_data = cpl_image_get_data_float(big_image);

	for (j=0; j<ys ; j++){
		inrow = j*xs;
		outrow = (j+dy)*xx;

		for (i=0; i<xs ; i++){
			out_data[outrow+dx+i] = data[inrow+i];
		}

		for (i=0; i<dx; i++){
			out_data[outrow+i] = data[inrow+dx-i-1];
			out_data[outrow+xs+dx+i] = data[inrow+xs-i-1];

		}
	}

	for (j=0; j<dy ; j++) {

		for (i=0; i<xx; i++) {
			out_data[j*xx+i] = out_data[(2*dy-j-1)*xx+i];
			out_data[(yy-j-1)*xx+i] = out_data[(yy-2*dy+j)*xx+i];
		}
	}

	return big_image;
}


/**@}*/
