/* $Id: omega_mbias.c,v 1.3 2012-08-30 07:46:52 agabasch Exp $
 *
 * This file is part of the OMEGA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: agabasch $
 * $Date: 2012-08-30 07:46:52 $
 * $Revision: 1.3 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include "omega_recipe.h"

/**
 * @defgroup omega_mbias  Master Bias
 *
 * @par Synopsis:
 * Create a Master Bias for each chip. Trims and applies overscan correction to the raw input \n
 * bias frames. Averages these frames to derive the master bias frame.Calculates the image statistics on the \n
 * resulting frame. The recipes expects the following tags in the Set of Frames:
 *
 * @par Input frames:\n
 * BIAS\n
 * REFERENCE_BIAS (optional)
 *
 * @par Output frames:\n
 * MASTER_BIAS
 *
 *
 * @code
 *      #include omega_recipe.h
 * @endcode
 */
/**@{*/
/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int omega_mbias_create(cpl_plugin *) ;
static int omega_mbias_exec(cpl_plugin *) ;
static int omega_mbias_destroy(cpl_plugin *) ;
static int omega_mbias(cpl_frameset *,cpl_parameterlist *) ;

/*-----------------------------------------------------------------------------
                            Private Functions
 -----------------------------------------------------------------------------*/
int omega_mbias_combine(int xn);
static void omega_mbias_init(void);
static void omega_mbias_tidy(void);

/*-----------------------------------------------------------------------------
                            Static structures
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs. Parameters */
    float       rejsig;
    double       sigma;
    int         extnum;
    int         oc;
    int         paf;

    /* Outputs. QC parameters */
    int		CountHotPixels;
    double	Mean;
    double	Median;
    double 	Stdev;

}omega_mbias_config;


/* Input and Output */
static struct {
    cpl_size        *labels;
    const cpl_frame *rnframe;
    cpl_frameset     *biaslist;
    cpl_stats        *stats;
    omega_fits *firstbias;
    cpl_propertylist *ph;
    cpl_propertylist *eh;

    /* Products */
    cpl_image *mbias;
    cpl_image *hpixels;

} ps;

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

/* Static variables */
#define RECIPE "omega_mbias"

static int isfirst;
static int dummy;



/*!----------------------------------------------------------------------------

  @internal
  @brief    Build the list of available plugins, for this module.
  @param    list    the plugin list
  @return   0 if everything is ok

  This function is exported.

----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist * list)
{
    cpl_recipe  *   recipe = cpl_calloc(1, sizeof(*recipe)) ;
    cpl_plugin  *   plugin = &recipe->interface ;

    cpl_plugin_init(plugin,
                    CPL_PLUGIN_API,
                    OMEGA_BINARY_VERSION,
                    CPL_PLUGIN_TYPE_RECIPE,
                    "omega_mbias",
                    "OMEGA - Create Master Bias for each chip.",
                    "Trims and applies overscan correction to the raw input bias \n"
           	"frames. Averages these frames to derive the master bias frame. \n"
		"Calculates the image statistics on the resulting frame. Creates \n"
		"the FITS header and saves it together with the FITS image.",
                    "Sandra Castro",
                    "scastro@eso.org",
                    omega_get_license(),
                    omega_mbias_create,
                    omega_mbias_exec,
                    omega_mbias_destroy) ;

    cpl_pluginlist_append(list, plugin) ;

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options
  @param    plugin  the plugin
  @return   0 if everything is ok

  Create the recipe instance and make it available to the application using the
  interface.
 */
/*----------------------------------------------------------------------------*/
static int omega_mbias_create(cpl_plugin * plugin)
{
    cpl_recipe * recipe;
    cpl_parameter * p ;

    /* Do not create the recipe if an error code is already set */
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());
        return (int)cpl_error_get_code();
    }

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    /* Create the parameters list in the cpl_recipe object */
    recipe->parameters = cpl_parameterlist_new() ;
    if (recipe->parameters == NULL) {
        cpl_msg_error(cpl_func, "Parameter list allocation failed");
        cpl_ensure_code(0, (int)CPL_ERROR_ILLEGAL_OUTPUT);
    }

    /* Fill the parameters list */
   p = cpl_parameter_new_value("omega.omega_mbias.ExtensionNumber",
                                CPL_TYPE_INT,
                               "FITS extension number to load (1 to 32). (-1 = all)",
                                "omega_mbias",
                                -1) ;

    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"ext") ;
    cpl_parameterlist_append(recipe->parameters, p) ;


    p = cpl_parameter_new_range("omega.omega_mbias.OverscanMethod",
                                    CPL_TYPE_INT,
                                    "Overscan Correction Method",
                                    "omega_mbias",
                                    0, 0, 6);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"oc-meth") ;
    cpl_parameterlist_append(recipe->parameters, p) ;

    p = cpl_parameter_new_value("omega.omega_mbias.PAF",
                                CPL_TYPE_BOOL,
                                "Boolean value to create PAF files. 1(Yes), 0(No)",
                                "omega_mbias",
                                1) ;

    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "paf") ;
    cpl_parameterlist_append(recipe->parameters, p) ;

    p = cpl_parameter_new_value("omega.omega_mbias.SigmaClip",
                                CPL_TYPE_DOUBLE,
                                "Sigma Clipping Threshold for image",
                                "omega_mbias",
                                3.0) ;

    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"sig-clip") ;
    cpl_parameterlist_append(recipe->parameters, p) ;

    p = cpl_parameter_new_range("omega.omega_mbias.RejSigma",
                                CPL_TYPE_DOUBLE,
                                "Rejection threshold for outlying pixels in Hot Pixels Map",
                                "omega_mbias",
                                5.0, 1.0, 10.0) ;

    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "rej-sigma") ;
    cpl_parameterlist_append(recipe->parameters, p) ;


    /* Return */
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int omega_mbias_exec(cpl_plugin * plugin)
{
    cpl_recipe * recipe;
    int recipe_status;
/*   cpl_errorstate initial_errorstate = cpl_errorstate_get();*/

    /* Return immediately if an error code is already set */
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());
        return (int)cpl_error_get_code();
    }

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    /* Verify parameter and frame lists */
    if (recipe->parameters == NULL) {
        cpl_msg_error(cpl_func, "Recipe invoked with NULL parameter list");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }
    if (recipe->frames == NULL) {
        cpl_msg_error(cpl_func, "Recipe invoked with NULL frame set");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

     /* Invoke the recipe */
    recipe_status = omega_mbias(recipe->frames, recipe->parameters);

    /* Ensure DFS-compliance of the products */
    if (cpl_dfs_update_product_header(recipe->frames)) {
        if (!recipe_status) recipe_status = (int)cpl_error_get_code();
    }


   return recipe_status;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int omega_mbias_destroy(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe;

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    cpl_parameterlist_delete(recipe->parameters);

    return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here
  @param    set   the frames list
  @param    pars     the parameters list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int omega_mbias(cpl_frameset *set, cpl_parameterlist *pars)
{

    int j,jst,jfn,status;
    cpl_size nlab;
    int nbias = 0;
    char *outhpm = NULL;
    char *outmbias = NULL;

    const char *_id = "omega_mbias";

    cpl_parameter *par;
    cpl_frame *firstframe;
    cpl_frame *prframe_mbias = NULL;
    cpl_frame *prframe_hpm = NULL;
    const cpl_frame *refframe = NULL;
    cpl_propertylist    *qclist;
    cpl_propertylist    *alist;
    cpl_stats *diffstats = NULL;


 /*Start the recipe */

    if (!pars) {
        cpl_msg_error (_id, "Parameters list not found");
        return -1;
    }

    if (cpl_frameset_is_empty(set) == 1) {
        cpl_msg_error (_id, "Frameset not found");
        return -1;
    }


/* Retrieve input parameters */
    par = cpl_parameterlist_find(pars, "omega.omega_mbias.ExtensionNumber") ;
    omega_mbias_config.extnum = cpl_parameter_get_int(par) ;

    par = cpl_parameterlist_find(pars, "omega.omega_mbias.OverscanMethod") ;
    omega_mbias_config.oc = cpl_parameter_get_int(par) ;

    par = cpl_parameterlist_find(pars, "omega.omega_mbias.SigmaClip") ;
    omega_mbias_config.sigma = cpl_parameter_get_double(par) ;

    par = cpl_parameterlist_find(pars, "omega.omega_mbias.RejSigma") ;
    omega_mbias_config.rejsig = cpl_parameter_get_double(par) ;

    par = cpl_parameterlist_find(pars, "omega.omega_mbias.PAF");
    omega_mbias_config.paf = cpl_parameter_get_bool(par);



/* Identify the RAW and CALIB frames in the input frameset */
    if (oc_dfs_set_groups(set)) {
        cpl_msg_error(_id, "Cannot identify RAW and CALIB frames") ;
        return -1 ;
    }


    /*Initialized things*/
    omega_mbias_init();


    /* Verify the frameset contents. */
    if ((ps.labels = cpl_frameset_labelise(set,omega_compare_tags,
					   &nlab)) == NULL) {
        cpl_msg_error(_id,"Cannot labelise the input frameset");
        omega_mbias_tidy();
        return -1;
    }
    if ((ps.biaslist = omega_frameset_subgroup(set,ps.labels,nlab,
						MBIAS_RAW)) == NULL) {
        cpl_msg_error(_id,"Cannot find bias frames in input frameset");
        omega_mbias_tidy();
        return -1;
    }

    nbias = cpl_frameset_get_size(ps.biaslist);
    if (nbias < 2) {
        cpl_msg_error(_id,"Need at least 2 (%s) frames to run this recipe",MBIAS_RAW);
        omega_mbias_tidy();
        return -1;
    }

    cpl_msg_info (_id,"There are %d %s in frame set",nbias, MBIAS_RAW);

    /* Check for calibration frame */
    /* Readnoise table */
    ps.rnframe = cpl_frameset_find_const(set, OMEGA_CALIB_RDNOISE);
    if (ps.rnframe == NULL) {
        cpl_msg_info(_id,"A Readnoise table is not present in frame set. Using header values");
    }
    else{
        cpl_msg_info(_id,"Using %s %s",OMEGA_CALIB_RDNOISE, cpl_frame_get_filename(ps.rnframe));
    }

    /* Get first frame from frame set */
    firstframe = cpl_frameset_get_first(ps.biaslist);

    /* Optional reference master bias frame */
    refframe = cpl_frameset_find_const(set, REFBIAS);
    if (refframe != NULL)
        cpl_msg_info(cpl_func,"Using %s for comparison",cpl_frame_get_filename(refframe));

    /* Loop for each of the image extensions */
    omega_extensions(firstframe,omega_mbias_config.extnum,&jst,&jfn);

    if(omega_mbias_config.extnum == 0){
    	cpl_msg_error(cpl_func,"Unsupported extension request, %d",omega_mbias_config.extnum);
    	omega_mbias_tidy();
    	return -1;
    }


    /* Check which instrument */
    if (omega_pfits_check_instrume(firstframe) == 1)
         sprintf(INSTRUME,"wfi");

    if(omega_pfits_check_instrume(firstframe) == 1 &&
       omega_mbias_config.extnum == 0 && jfn == 32){
        jfn = 8;
    }

    for (j = jst; j <= jfn; j++) {

        isfirst = (j == jst);
        cpl_msg_info(_id,"Beginning work on extension %d",j);
        omega_mbias_config.CountHotPixels = 0;
        omega_mbias_config.Mean = 0.0;
        omega_mbias_config.Median = 0.0;
        omega_mbias_config.Stdev = 0.0;

        ps.firstbias = omega_fits_load(firstframe,CPL_TYPE_FLOAT,j);
        ps.eh = omega_fits_get_ehu(ps.firstbias);

        /* Call the combining routine */
        status = omega_mbias_combine(j);
        if(status == 1){
            cpl_msg_warning(_id, "Image detector is not live");
            /* Save dummy product */
            freefits(ps.firstbias);
            ps.eh = NULL;
            continue;
        }
        else if(status == -1){
            cpl_msg_error(_id,"Cannot combine images");
            freespace(outhpm);
            freespace(outmbias);
            omega_mbias_tidy();
            return -1;
        }

        status = 0;

         qclist = cpl_propertylist_new();

        /* Compare with reference frame */
        if(refframe != NULL){
            if(omega_compare_reference(ps.mbias,refframe,j,&diffstats) == -1){
                cpl_msg_warning(cpl_func,"Cannot compare with reference frame");
            }
            else{
                cpl_propertylist_append_double(qclist,"ESO QC DIFF REFBIAS MEAN",
                                               cpl_stats_get_mean(diffstats));
                cpl_propertylist_set_comment(qclist,"ESO QC DIFF REFBIAS MEAN",
                                             "Mean of difference with reference");
                cpl_propertylist_append_double(qclist,"ESO QC DIFF REFBIAS MEDIAN",
                                               cpl_stats_get_median(diffstats));
                cpl_propertylist_set_comment(qclist,"ESO QC DIFF REFBIAS MEDIAN",
                                             "Median of difference with reference");
                cpl_propertylist_append_double(qclist,"ESO QC DIFF REFBIAS STDEV",
                                               cpl_stats_get_stdev(diffstats));
                cpl_propertylist_set_comment(qclist,"ESO QC DIFF REFBIAS STDEV",
                                             "Stdev of difference with reference");

                freestats(diffstats);
            }
        }

        /* Save the products */
        /* Save the MASTER BIAS product */
        cpl_propertylist_append_double(qclist, "ESO QC MASTER BIAS MEAN",
                                    omega_mbias_config.Mean) ;
        cpl_propertylist_set_comment (qclist, "ESO QC MASTER BIAS MEAN",
                                      "Mean value of master bias");

        cpl_propertylist_append_double(qclist, "ESO QC MASTER BIAS MEDIAN",
                                    omega_mbias_config.Median) ;
        cpl_propertylist_set_comment (qclist, "ESO QC MASTER BIAS MEDIAN",
                                      "Median value of master bias");

        cpl_propertylist_append_double(qclist, "ESO QC MASTER BIAS STDEV",
                                    omega_mbias_config.Stdev) ;
        cpl_propertylist_set_comment (qclist, "ESO QC MASTER BIAS STDEV",
                                      "Standard Deviation of master bias");

		if(isfirst){
			outmbias = cpl_sprintf("%s_%s.fits", INSTRUME,MBIAS_PROCATG);
			prframe_mbias = omega_product_frame(outmbias, MBIAS_PROCATG, CPL_FRAME_TYPE_IMAGE);
		}

	       alist = cpl_propertylist_new();
	        cpl_propertylist_append_string(alist, "EXTNAME",
						cpl_propertylist_get_string(ps.eh, "EXTNAME"));
				cpl_propertylist_set_comment(alist,"EXTNAME", "Extension name");

				/* Add DRS keywords */
				cpl_propertylist_update_int(alist, "ESO DRS OVERSCAN METHOD", omega_mbias_config.oc);
				cpl_propertylist_set_comment(alist, "ESO DRS OVERSCAN METHOD", "overscan correction method");

				cpl_propertylist_copy_property_regexp(alist, ps.eh, WCS_KEYS, 0);


		if(omega_save_image(ps.mbias,set,pars,alist,qclist,CPL_BPP_IEEE_FLOAT,outmbias,
					RECIPE,prframe_mbias,NULL,isfirst) == -1){
            cpl_msg_error(_id,"Cannot save product %s", MBIAS_PROCATG);
            freeplist(qclist);
            freeplist(alist);
            freespace(outhpm);
            freespace(outmbias);
            omega_mbias_tidy();
            return -1;;
		}

        freeplist(qclist);
        freeimage(ps.mbias);

        /* Save the HOT PIXELS MAP product */
        qclist = cpl_propertylist_new();
        cpl_propertylist_append_int(qclist,"ESO QC NUMBER HOT PIXELS",
                                omega_mbias_config.CountHotPixels);
        cpl_propertylist_set_comment(qclist,"ESO QC NUMBER HOT PIXELS",
                                     "Number of hot pixels");
		if(isfirst){
			outhpm = cpl_sprintf("%s_%s.fits", INSTRUME,HPM_PROCATG);
			prframe_hpm = omega_product_frame(outhpm, HPM_PROCATG, CPL_FRAME_TYPE_IMAGE);
		}

		if(omega_save_image(ps.hpixels,set,pars,alist,qclist,CPL_BPP_16_SIGNED,outhpm,
					RECIPE,prframe_hpm,NULL,isfirst) == -1){
            cpl_msg_error(_id,"Cannot save product %s", HPM_PROCATG);
            freeplist(qclist);
            freeplist(alist);
           freespace(outhpm);
            freespace(outmbias);
            omega_mbias_tidy();
            return -1;;
		}

        freeplist(qclist);
        freeplist(alist);
        freeimage(ps.hpixels);
        status = 0;

        freefits(ps.firstbias);
        ps.firstbias = NULL;
        ps.eh = NULL;


    }  /* work on next extension */

    freespace(outhpm);
    freespace(outmbias);
    omega_mbias_tidy();

    return 0;
}

/**
 * @brief
 *   Process the BIAS frames
 *
 * @param xn  FITS extension to load
 * @return
 *   0 on success or -1 in case of error.
 */
int omega_mbias_combine(int xn)
{

    int i, nbias, live, naxis1, naxis2;
    double read_noise, threshold;
    double cutoff = 200.0; /*FIXME: arbitrary initial value*/
    const char *_id = "omega_mbias_combine";

    const cpl_frame *biasfr;
    cpl_image *trim_raw,*median_all,*dev,*good_int,*good_float;
    cpl_image *sum_data,*sum_good,*image;
    cpl_imagelist *ilist;
    cpl_mask *good,*pixelmap;
    cpl_table *rntable;


    nbias = cpl_frameset_get_size(ps.biaslist);
    biasfr = cpl_frameset_get_first_const(ps.biaslist);


    /* Check that this detector is live*/
    omega_pfits_get_detlive(ps.eh,&live);
    if (! live) {
        return 1;
     }

    cpl_msg_info (_id,"Doing trim and overscan correction on images");

    for (i=0; i< nbias; i++){

        trim_raw = TrimOscanCorrect(biasfr, omega_mbias_config.oc, xn);
        if(trim_raw == NULL){
            return -1;
        }

        if (i == 0) {
            naxis1 = cpl_image_get_size_x (trim_raw);
            naxis2 = cpl_image_get_size_y (trim_raw);
            ilist = cpl_imagelist_new ();
        }

        cpl_imagelist_set (ilist, trim_raw, i);

        biasfr = cpl_frameset_get_next_const(ps.biaslist);
        if (biasfr == NULL)
            break;

    }

    if (ilist == NULL) {
        cpl_msg_error(_id,"Error in imagelist <%s>",cpl_error_get_message());
        freeimage(trim_raw);
        return -1;

    }

    /*
     * If READNOISE_ADU table doesn't exist, try to
     * get readnoise value from header of first image
     */
    if(ps.rnframe != NULL){
    	rntable = cpl_table_load(cpl_frame_get_filename(ps.rnframe), xn, 0);
    	if(rntable != NULL){
     		read_noise = cpl_table_get_double(rntable, "READNOISE",0, NULL);
    		freetable(rntable);
    	}
    	else{
    		omega_pfits_get_readnoise(ps.eh, &read_noise);
    		cpl_msg_warning(cpl_func,"Cannot load READNOISE_ADU table. %s",cpl_error_get_message());
    	}
    }
    else {
    	omega_pfits_get_readnoise(ps.eh, &read_noise);
    }

    if (read_noise == 0.0){
        cpl_msg_info(_id,"Using default value for read noise");
        read_noise = 2.5;
    }
    else {
        cpl_msg_info(_id,"Using read noise = %g", read_noise);
    }


    threshold = read_noise * omega_mbias_config.sigma;
    cpl_msg_info (_id,"Discarding outliers using threshold %g", threshold);

/* Calculate the median of all raw bias */

    cpl_msg_info (_id,"Computing the median of all images...");
    median_all = cpl_imagelist_collapse_median_create (ilist);

    if (median_all == NULL) {
        cpl_msg_error (_id,"Cannot take median of imagelist <%s>",cpl_error_get_message());
        freeilist(ilist);
        return -1;
    }

/* Clip the outliers from the first image*/
    trim_raw = cpl_imagelist_get (ilist, 0);
    dev = cpl_image_subtract_create(median_all, trim_raw);

    good = cpl_mask_threshold_image_create(dev, -threshold, threshold);
    freeimage(dev);

    good_int = cpl_image_new_from_mask(good) ;
    freemask(good) ;
    good_float = cpl_image_cast(good_int, CPL_TYPE_FLOAT);
    freeimage(good_int);

    sum_data = cpl_image_multiply_create(trim_raw, good_float);
    sum_good = cpl_image_duplicate(good_float);

    freeimage(good_float);

/* Clip the outliers from the remaining images*/
    for (i=1; i< nbias; i++){

        trim_raw = cpl_imagelist_get(ilist, i);
/*			if (trim_raw == NULL)
				cpl_msg_error("","Image is NULL <%s>",cpl_error_get_message());
*/
        dev = cpl_image_subtract_create(median_all, trim_raw);

        if (dev == NULL) {
            cpl_msg_error(_id,"Error in subtraction <%s>",cpl_error_get_function());
            freeilist(ilist);
            freeimage(median_all);
            freeimage(sum_data);
            freeimage(sum_good);
            return -1;
        }

        good = cpl_mask_threshold_image_create (dev, -threshold, threshold);

        freeimage(dev);

        good_int = cpl_image_new_from_mask(good) ;
        freemask(good) ;
        good_float = cpl_image_cast(good_int, CPL_TYPE_FLOAT);
        freeimage(good_int);

        image = cpl_image_multiply_create(trim_raw, good_float);
        cpl_image_add(sum_data, image);

        cpl_image_add(sum_good, good_float);

        freeimage(image);
        freeimage(good_float);
    }

    freeimage(median_all);
    freeilist(ilist);

    ps.mbias = cpl_image_divide_create(sum_data, sum_good);

    if (ps.mbias == NULL) {
        cpl_msg_error(_id,"Cannot create product. Error in division");
        freeimage(sum_data);
        freeimage(sum_good);
        return -1;
    }

    freeimage(sum_data);
    freeimage(sum_good);

    /* Compute statistics on the master bias image */
    ps.stats = cpl_stats_new_from_image(ps.mbias, CPL_STATS_ALL);
    if(ps.stats != NULL) {
        omega_mbias_config.Mean = cpl_stats_get_mean(ps.stats);
        omega_mbias_config.Median = cpl_stats_get_median(ps.stats);
        omega_mbias_config.Stdev = cpl_stats_get_stdev(ps.stats);
        cutoff = omega_mbias_config.Mean + omega_mbias_config.rejsig * omega_mbias_config.Stdev;
        cpl_msg_info(_id,"Using cutoff %g for Hot Pixels Map creation", cutoff);
    }

    freestats(ps.stats);

    /* Create a hot pixels map */
    pixelmap = cpl_mask_threshold_image_create(ps.mbias, -1e20, cutoff);

    /*Do an inverse to count the number of CPL_BINARY_1*/
    cpl_mask_not(pixelmap);
    omega_mbias_config.CountHotPixels = cpl_mask_count(pixelmap);

    /*Invert back*/
    /*    cpl_mask_not(pixelmap);*/

    /* Convert image to INT to save it */

    ps.hpixels = cpl_image_new_from_mask(pixelmap);
    freemask(pixelmap);

    return 0;
}

/* Initialize the pointers */
static void omega_mbias_init(void) {
    ps.labels = NULL;
    ps.rnframe = NULL;
    ps.biaslist = NULL;
    ps.mbias = NULL;
    ps.hpixels = NULL;
    ps.ph = NULL;
    ps.eh = NULL;
    ps.stats = NULL;
    ps.firstbias = NULL;
}

/* Free any allocated memory */
static void omega_mbias_tidy(void) {
    freespace(ps.labels);
    freeframeset(ps.biaslist);
    freeimage(ps.mbias);
    freeimage(ps.hpixels);
    freestats(ps.stats);
    freefits(ps.firstbias);
}

/**@}*/
