/* $Id: omega_mdome.c,v 1.7 2012-08-30 07:46:52 agabasch Exp $
 *
 * This file is part of the OMEGA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: agabasch $
 * $Date: 2012-08-30 07:46:52 $
 * $Revision: 1.7 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include "omega_recipe.h"
#include "omega_background.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup omega_mdome  Master Dome Flat
 *
 * @par Synopsis:
 * This recipe is used to derive a master dome flat for one particular chip and filter. The recipe always \n
 * takes as input a list of raw dome flat frames, a master bias and optionals hot pixel and a cold\n
 * pixel maps. Optionally, an overscan correction mode can be set. The default is to apply no overscan \n
 * correction. The raw dome flats are trimmed and overscan corrected. The data are then normalized, averaged, \n
 * and the result is normalized again. Image statistics are computed for the resulting frames. A new cold\n
 * pixels map is created from the flats. The recipe expects the following tags in the Set of Frames:
 *
 * @par Input frames:
 * DOME_FLAT\n
 * MASTER_BIAS\n
 * HOT_PIXELS_MAP\n
 * COLD_PIXELS_MAP (optional)\n
 * REFERENCE_DOME (optional)
 *
 * @par Output frames:
 * MASTER_FLAT_DOME\n
 * COLD_PIXELS_MAP
 *
 * @code
 *      #include omega_recipe.h
 *      #include omega_background.h
 * @endcode
 */
/**@{*/


/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int omega_mdome_create(cpl_plugin *);
static int omega_mdome_exec(cpl_plugin *);
static int omega_mdome_destroy(cpl_plugin *);
static int omega_mdome(cpl_frameset *, cpl_parameterlist *);

/*-----------------------------------------------------------------------------
                            Private Functions
 -----------------------------------------------------------------------------*/

int omega_mdome_combine(cpl_parameterlist *pars, int xn);
static void omega_mdome_init(void);
static void omega_mdome_tidy(void);

/*-----------------------------------------------------------------------------
                            Static structures
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs. Parameters */
    int         extnum;
    int         oc;
    int         paf;
    double      sigma;
    double lthre;
    double hthre;

    /* Outputs. QC parameters */
    int		CountColdPixels;
    double	Mean;
    double	Median;
    double 	Stdev;
    double	RawMin;
    double	RawMax;
    double	RawMean;
    double	RawMedian;
    double	RawStdev;

}omega_mdome_config;

/* Input and Output */
static struct {

    cpl_size         *labels;
    const cpl_frame  *mbframe;
    const cpl_frame  *cframe;
    const cpl_frame  *hframe;
    cpl_frameset     *domelist;
    cpl_stats        *stats;
    cpl_propertylist *ph;
    cpl_propertylist *eh;
    omega_fits *firstdome;

    /* Products */
    cpl_image *mdome;
    cpl_image *cpixels;

}ps;

/*-----------------------------------------------------------------------------
                            Static global variables
 -----------------------------------------------------------------------------*/

/* Static variables */
#define RECIPE "omega_mdome"

static int isfirst;


/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module.
  @param    list    the plugin list
  @return   0 if everything is ok

  This function is exported.
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist * list)
{
    cpl_recipe  *   recipe = cpl_calloc(1, sizeof(*recipe)) ;
    cpl_plugin  *   plugin = &recipe->interface ;

    cpl_plugin_init(plugin,
                    CPL_PLUGIN_API,
                    OMEGA_BINARY_VERSION,
                    CPL_PLUGIN_TYPE_RECIPE,
                    "omega_mdome",
                    "OMEGA - Create Master Dome Flat for each chip (Calfile 542).",
                    "This recipe is used to derive a valid domeflat frame (Calfile 542) \n"
                    "for one particular chip and filter. The recipe always takes as input \n"
                    "a list of raw domeflat frames, a master bias and a measurement of the \n"
                    "gain of the chip under consideration. In addition, a hot pixel map \n"
                    "(Calfile 522) and a cold pixel map (Calfile 535) can be provided. Optionally, \n"
                    "an overscan correction mode can be set. The default is to apply no overscan \n"
                    "correction.\n\n"
                    "The raw dome flats are trimmed and overscan corrected. The data are then \n"
                    "normalized, averaged, and the result is normalized again. Image statisics \n"
                    "are computed for the resulting frame.",
                    "Sandra Castro",
                    "scastro@eso.org",
                    omega_get_license(),
                    omega_mdome_create,
                    omega_mdome_exec,
                    omega_mdome_destroy) ;

    cpl_pluginlist_append(list, plugin) ;

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options
  @param    plugin  the plugin
  @return   0 if everything is ok

  Create the recipe instance and make it available to the application using the
  interface.
 */
/*----------------------------------------------------------------------------*/
static int omega_mdome_create(cpl_plugin * plugin)
{
    cpl_recipe * recipe;
    cpl_parameter * p ;
    char *path = NULL;

    /* Do not create the recipe if an error code is already set */
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());
        return (int)cpl_error_get_code();
    }

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    /* Create the parameters list in the cpl_recipe object */
    recipe->parameters = cpl_parameterlist_new() ;

    if (recipe->parameters == NULL) {
        cpl_msg_error(cpl_func, "Parameter list allocation failed");
        cpl_ensure_code(0, (int)CPL_ERROR_ILLEGAL_OUTPUT);
    }

    /* Fill the parameters list */

    /* General Parameters */
   p = cpl_parameter_new_value("omega.omega_mdome.ExtensionNumber",
                                CPL_TYPE_INT,
                               "FITS extension number to load (1 to 32). (-1 = all)",
                                "omega_mdome",
                                -1) ;

    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"ext") ;
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p) ;


    p = cpl_parameter_new_range("omega.omega_mdome.OverscanMethod",
                                    CPL_TYPE_INT,
                                    "Overscan Correction Method",
                                    "omega_mdome",
                                    0, 0, 6);

    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"oc-meth") ;
    cpl_parameterlist_append(recipe->parameters, p) ;

    p = cpl_parameter_new_value("omega.omega_mdome.PAF",
                                CPL_TYPE_BOOL,
                                "Boolean value to create PAF files. 1(Yes), 0(No)",
                                "omega_mdome",
                                1) ;

    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "paf") ;
    cpl_parameterlist_append(recipe->parameters, p) ;

    p = cpl_parameter_new_value("omega.omega_mdome.SigmaClip",
                                CPL_TYPE_DOUBLE,
                                "Sigma Clipping Threshold",
                                "omega_mdome",
                                3.0) ;

    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "sig-clip") ;
    cpl_parameterlist_append(recipe->parameters, p) ;


/*------------ Parameters for creating a Cold Pixels Map---------------- */
    p = cpl_parameter_new_range("omega.omega_mdome.LowThre",
                                CPL_TYPE_DOUBLE,
                                "Low flagging threshold for cold pixels map",
                                "omega_coldpixels",
                                0.93, 0.90, 1.00) ;

    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "low") ;
    cpl_parameterlist_append(recipe->parameters, p) ;

    p = cpl_parameter_new_range("omega.omega_mdome.HighThre",
                                CPL_TYPE_DOUBLE,
                                "High flagging threshold for cold pixels map",
                                "omega_coldpixels",
                                1.07, 1.00, 1.10) ;

    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"high") ;
    cpl_parameterlist_append(recipe->parameters, p) ;

    p = cpl_parameter_new_range("omega.omega_mdome.BackSize",
                                CPL_TYPE_INT,
                                "Sextractor background mesh size for cold pixels map",
                                "omega_coldpixels",
                                32, 4, 256) ;

    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"backsize") ;
    cpl_parameterlist_append(recipe->parameters, p) ;

/* ------------------ PATH to external executable programs --------------*/
    path = cpl_sprintf("%s", OMEGA_BIN_PATH);
    p = cpl_parameter_new_value("omega.omega_mdome.BinPath",
                                CPL_TYPE_STRING,
                                "Path to any external executable program.",
                                "omega.BinPath",
                                path);

    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "bin-path");
    cpl_parameterlist_append(recipe->parameters, p);
    cpl_free(path);

/* ------------------ Sextractor parameters ------------------------- */
    path = cpl_sprintf("%s/omega.sex", OMEGA_CONFIG_PATH);
    p = cpl_parameter_new_value("omega.omega_mdome.SexConfig",
                                CPL_TYPE_STRING,
                                "Path to Sextractor config file.",
                                "omega.Sextractor",
                                path);

    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "sex-config");
    cpl_parameterlist_append(recipe->parameters, p);
    cpl_free(path);

    path = cpl_sprintf("%s/omega.conv", OMEGA_CONFIG_PATH);
    p = cpl_parameter_new_value("omega.omega_mdome.SexConv",
                                CPL_TYPE_STRING,
                                "Path to Sextractor convolution mask file.",
                                "omega.Sextractor",
                                path);

    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"sex-conv");
    cpl_parameterlist_append(recipe->parameters, p);
    cpl_free(path);


    path = cpl_sprintf("%s/omega.param", OMEGA_CONFIG_PATH);
    p = cpl_parameter_new_value("omega.omega_mdome.SexParam",
                                CPL_TYPE_STRING,
                                "Path to Sextractor parameters file.",
                                "omega.Sextractor",
                                path);

    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "sex-param");
    cpl_parameterlist_append(recipe->parameters, p);
    cpl_free(path);

    path = cpl_sprintf("%s/omega.nnw", OMEGA_CONFIG_PATH);
    p = cpl_parameter_new_value("omega.omega_mdome.SexNnw",
                                CPL_TYPE_STRING,
                                "Path to Sextractor neural network config file.",
                                "omega.Sextractor",
                                path);

    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"sex-nnw");
    cpl_parameterlist_append(recipe->parameters, p);
    cpl_free(path);

    p = cpl_parameter_new_value("omega.omega_mdome.BackThreshold",
                                CPL_TYPE_DOUBLE,
                                "Detection threshold for background in image (Sextractor DETECT_THRESH)",
                                "omega_Sextractor",
                                1000.) ;

    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "backthre") ;
    cpl_parameterlist_append(recipe->parameters, p) ;



    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int omega_mdome_exec(cpl_plugin * plugin)
{
    cpl_recipe * recipe;
    int recipe_status;
/*    cpl_errorstate initial_errorstate = cpl_errorstate_get();*/

    /* Return immediately if an error code is already set */
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());
        return (int)cpl_error_get_code();
    }

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    /* Verify parameter and frame lists */
    if (recipe->parameters == NULL) {
        cpl_msg_error(cpl_func, "Recipe invoked with NULL parameter list");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }
    if (recipe->frames == NULL) {
        cpl_msg_error(cpl_func, "Recipe invoked with NULL frame set");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Invoke the recipe */
    recipe_status = omega_mdome(recipe->frames, recipe->parameters);

    /* Ensure DFS-compliance of the products */
    if (cpl_dfs_update_product_header(recipe->frames)) {
        if (!recipe_status) recipe_status = (int)cpl_error_get_code();
    }

        /* Dump the error history since recipe execution start.
           At this point the recipe cannot recover from the error */
    /*   if (!cpl_errorstate_is_equal(initial_errorstate)) {
        cpl_errorstate_dump(initial_errorstate, CPL_FALSE, NULL);
    }*/

    return recipe_status;

}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int omega_mdome_destroy(cpl_plugin * plugin)
{
    cpl_recipe  *recipe;

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    cpl_parameterlist_delete(recipe->parameters);

    return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here
  @param    parlist     the parameters list
  @param    framelist   the frames list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int omega_mdome(cpl_frameset *set, cpl_parameterlist *pars)
{

    int j,jst,jfn;
    cpl_size nlab;
    int nflats = 0;
    int oscan1 = 0;
    int status = 1;

    float bias = BIAS;

    char *outcpm = NULL;
    char *outmdome = NULL;
    const char *_id = "omega_mdome";
    cpl_frame 		*firstframe = NULL;
    cpl_frame           *prframe_cpm = NULL;
    cpl_frame           *prframe_mdome = NULL;
    const cpl_frame     *refframe = NULL;
    cpl_parameter 	*par = NULL;
    cpl_propertylist    *qclist = NULL;
    cpl_propertylist    *alist = NULL;
    cpl_stats           *diffstats = NULL;


 /*Start the recipe */

    if (!pars) {
        cpl_msg_error (_id, "Parameters list not found");
        return -1;
    }

    if (cpl_frameset_is_empty(set) == 1) {
        cpl_msg_error (_id, "Frameset not found");
        return -1;
    }

/* Retrieve input parameters */
    par = cpl_parameterlist_find(pars, "omega.omega_mdome.ExtensionNumber") ;
    omega_mdome_config.extnum = cpl_parameter_get_int(par) ;

    par = cpl_parameterlist_find(pars, "omega.omega_mdome.OverscanMethod") ;
    omega_mdome_config.oc = cpl_parameter_get_int(par) ;

    par = cpl_parameterlist_find(pars, "omega.omega_mdome.SigmaClip") ;
    omega_mdome_config.sigma = cpl_parameter_get_double(par) ;

    par = cpl_parameterlist_find(pars, "omega.omega_mdome.LowThre") ;
    omega_mdome_config.lthre = cpl_parameter_get_double(par) ;

    par = cpl_parameterlist_find(pars, "omega.omega_mdome.HighThre") ;
    omega_mdome_config.hthre = cpl_parameter_get_double(par) ;

    par = cpl_parameterlist_find(pars, "omega.omega_mdome.PAF") ;
    omega_mdome_config.paf = cpl_parameter_get_bool(par) ;

/* Identify the RAW and CALIB frames in the input frameset */
    if (oc_dfs_set_groups(set)) {
        cpl_msg_error(_id, "Cannot identify RAW and CALIB frames") ;
        return -1 ;
    }

    /*Initialized things*/
    omega_mdome_init();

/* Verify the frameset contents. */
    if ((ps.labels = cpl_frameset_labelise(set,omega_compare_tags,
					   &nlab)) == NULL) {
        cpl_msg_error(_id,"Cannot labelise the input frameset");
        omega_mdome_tidy();
        return -1;
    }
    if ((ps.domelist = omega_frameset_subgroup(set,ps.labels,nlab,
						MDOME_RAW)) == NULL) {
        cpl_msg_error(_id,"Cannot find dome frames in input frameset");
        omega_mdome_tidy();
        return -1;
    }


    nflats = cpl_frameset_count_tags(ps.domelist, MDOME_RAW);
    if (nflats < 2) {
        cpl_msg_error (_id, "Need at least 2 (%s) frames to run "
                            "this recipe", MDOME_RAW);
        omega_mdome_tidy();
        return -1;
    }

    cpl_msg_info (_id,"There are %d %s in frame set",nflats, MDOME_RAW);

    /* Check for calibration frames */
    /* Master Bias */
    ps.mbframe = cpl_frameset_find_const(set, OMEGA_CALIB_BIAS);
    if (ps.mbframe == NULL) {
        cpl_msg_info(_id,"A Master Bias is not present in frame set. Using default value %f", bias);
    }
    else{
        cpl_msg_info(_id,"Using %s %s",OMEGA_CALIB_BIAS, cpl_frame_get_filename(ps.mbframe));
    }

    /* Cold Pixels Map */
    ps.cframe = cpl_frameset_find_const(set, OMEGA_CALIB_CPM);
    if(ps.cframe != NULL){
        cpl_msg_info(_id,"Using %s %s",OMEGA_CALIB_CPM, cpl_frame_get_filename(ps.cframe));
    }

    /* Hot Pixels Map */
    ps.hframe = cpl_frameset_find_const(set, OMEGA_CALIB_HPM);
    if(ps.hframe != NULL){
        cpl_msg_info(_id,"Using %s %s",OMEGA_CALIB_HPM, cpl_frame_get_filename(ps.hframe));
    }

    /* Optional Reference master dome frame */
    refframe = cpl_frameset_find_const(set, REFDOME);
    if (refframe != NULL)
        cpl_msg_info(cpl_func,"Using %s for comparison",cpl_frame_get_filename(refframe));

    /* Get first frame from frame set */
    firstframe = cpl_frameset_get_first(ps.domelist);

    /* Loop for each of the image extensions */
    omega_exten_range(omega_mdome_config.extnum,&jst,&jfn);
    if(omega_mdome_config.extnum == 0){
    	cpl_msg_error(cpl_func,"Unsupported extension request, %d",omega_mdome_config.extnum);
    	omega_mdome_tidy();
    	return -1;
    }

    /* Check which instrument */
    if (omega_pfits_check_instrume(firstframe) == 1)
         sprintf(INSTRUME,"wfi");

    /* Check which extensions */
    if(omega_pfits_check_instrume(firstframe) == 1 &&
                  omega_mdome_config.extnum == 0 && jfn == 32)
        jfn = 8;

    for (j = jst; j <= jfn; j++) {

        isfirst = (j == jst);
        cpl_msg_info(_id,"Beginning work on extension %d",j);
        omega_mdome_config.CountColdPixels = 0;
        omega_mdome_config.Mean = 0.0;
        omega_mdome_config.Median = 0.0;
        omega_mdome_config.Stdev = 0.0;
        omega_mdome_config.RawMin = 0.0;
        omega_mdome_config.RawMax = 0.0;
        omega_mdome_config.RawMean = 0.0;
        omega_mdome_config.RawMedian = 0.0;
        omega_mdome_config.RawStdev = 0.0;

        /* Check overscan correction method consistency */
        if(ps.mbframe != NULL){
            oscan1 = omega_pfits_get_overscan(ps.mbframe, j);
            if(oscan1 != omega_mdome_config.oc) {
                cpl_msg_warning (_id, "Overscan correction mode for Master Bias (oc = %d) differs from "
                                        "the one used here (oc = %d)", oscan1, omega_mdome_config.oc);
            }
        }


        /* Load first dome flat */
        ps.firstdome = omega_fits_load(firstframe,CPL_TYPE_FLOAT,j);
        ps.eh = omega_fits_get_ehu(ps.firstdome);

        /* Call the combining routine */
        status = omega_mdome_combine(pars, j);
        if(status == 1){
            cpl_msg_warning(_id, "Image detector is not live");
            /* Save dummy product */
            freefits(ps.firstdome);
            ps.eh = NULL;
            continue;
        }
        else if(status == -1){
            cpl_msg_error(_id,"Cannot combine images");
            freespace(outcpm);
            freespace(outmdome);
            omega_mdome_tidy();
            return -1;
        }

         qclist = cpl_propertylist_new();

        /* Compare with reference frame */
        if(refframe != NULL){
            if(omega_compare_reference(ps.mdome,refframe,j,&diffstats) == -1){
                cpl_msg_warning(cpl_func,"Cannot compare with reference frame");
            }
            else{
                cpl_propertylist_append_double(qclist,"ESO QC DIFF REFDOME MEAN",
                                               cpl_stats_get_mean(diffstats));
                cpl_propertylist_set_comment(qclist,"ESO QC DIFF REFDOME MEAN",
                                             "Mean of difference with reference");
                cpl_propertylist_append_double(qclist,"ESO QC DIFF REFDOME MEDIAN",
                                               cpl_stats_get_median(diffstats));
                cpl_propertylist_set_comment(qclist,"ESO QC DIFF REFDOME MEDIAN",
                                             "Median of difference with reference");
                cpl_propertylist_append_double(qclist,"ESO QC DIFF REFDOME STDEV",
                                               cpl_stats_get_stdev(diffstats));
                cpl_propertylist_set_comment(qclist,"ESO QC DIFF REFDOME STDEV",
                                             "Stdev of difference with reference");

                freestats(diffstats);
            }
        }

        /* Save the products */
        /* Save the MASTER DOME product */

        cpl_propertylist_append_double(qclist, "ESO QC MASTER DOME MEAN",
                                            omega_mdome_config.Mean) ;
        cpl_propertylist_set_comment (qclist, "ESO QC MASTER DOME MEAN",
                                      "Mean of master dome flat");

        cpl_propertylist_append_double(qclist, "ESO QC MASTER DOME MEDIAN",
                                            omega_mdome_config.Median) ;
        cpl_propertylist_set_comment (qclist, "ESO QC MASTER DOME MEDIAN",
                                      "Median of master dome flat");

        cpl_propertylist_append_double(qclist, "ESO QC MASTER DOME STDEV",
                                            omega_mdome_config.Stdev) ;
        cpl_propertylist_set_comment (qclist, "ESO QC MASTER DOME STDEV",
                                      "Std Deviation of master dome flat");

        cpl_propertylist_append_double(qclist, "ESO QC RAW DOME MIN",
        		omega_mdome_config.RawMin);
        cpl_propertylist_append_double(qclist, "ESO QC RAW DOME MAX",
        		omega_mdome_config.RawMax);
        cpl_propertylist_append_double(qclist, "ESO QC RAW DOME MEAN",
        		omega_mdome_config.RawMean);
        cpl_propertylist_append_double(qclist, "ESO QC RAW DOME MEDIAN",
        		omega_mdome_config.RawMedian);
        cpl_propertylist_append_double(qclist, "ESO QC RAW DOME STDEV",
        		omega_mdome_config.RawStdev);

        cpl_propertylist_set_comment(qclist, "ESO QC RAW DOME MIN",
        		"median value of the raw dome flat having the lowest flux");
        cpl_propertylist_set_comment(qclist, "ESO QC RAW DOME MAX",
        		"median value of the raw dome flat having the highest flux");
        cpl_propertylist_set_comment(qclist, "ESO QC RAW DOME MEAN",
        		"mean value of all input raw dome flats (ADU)");
        cpl_propertylist_set_comment(qclist, "ESO QC RAW DOME MEDIAN",
        		"median value of all input raw dome flats (ADU)");
        cpl_propertylist_set_comment(qclist, "ESO QC RAW DOME STDEV",
        		"standard deviation of all input raw dome flats (ADU)");

        if(isfirst){
        	outmdome = cpl_sprintf("%s_%s.fits", INSTRUME,MDOME_PROCATG);
        	prframe_mdome = omega_product_frame(outmdome, MDOME_PROCATG, CPL_FRAME_TYPE_IMAGE);
        }

        alist = cpl_propertylist_new();
        cpl_propertylist_append_string(alist, "EXTNAME",
        		cpl_propertylist_get_string(ps.eh, "EXTNAME"));
        cpl_propertylist_set_comment(alist,"EXTNAME", "Extension name");

        /* Add DRS keywords */
		cpl_propertylist_update_int(alist, "ESO DRS OVERSCAN METHOD", omega_mdome_config.oc);
		cpl_propertylist_set_comment(alist, "ESO DRS OVERSCAN METHOD", "overscan correction method");

		cpl_propertylist_copy_property_regexp(alist, ps.eh, WCS_KEYS, 0);

        if(omega_save_image(ps.mdome,set,pars,alist,qclist,CPL_BPP_IEEE_FLOAT,outmdome,
        		RECIPE,prframe_mdome,NULL,isfirst) == -1){
        	cpl_msg_error(_id,"Cannot save product %s", MDOME_PROCATG);
        	freeplist(qclist);
        	freeplist(alist);
        	freespace(outcpm);
        	freespace(outmdome);
        	omega_mdome_tidy();
        	return -1;;
        }
        freeplist(qclist);
        freeimage(ps.mdome);

        /* Save the COLD PIXELS MAP product */
        qclist = cpl_propertylist_new();
        cpl_propertylist_append_int(qclist,"ESO QC NUMBER COLD PIXELS",
                                        omega_mdome_config.CountColdPixels);

        cpl_propertylist_set_comment(qclist,"ESO QC NUMBER COLD PIXELS",
                                     "Number of cold pixels");
		if(isfirst){
			outcpm = cpl_sprintf("%s_%s.fits", INSTRUME,CPM_PROCATG);
			prframe_cpm = omega_product_frame(outcpm, CPM_PROCATG, CPL_FRAME_TYPE_IMAGE);
		}

		if(omega_save_image(ps.cpixels,set,pars,alist,qclist,CPL_BPP_16_SIGNED,outcpm,
					RECIPE,prframe_cpm,NULL,isfirst) == -1){
            cpl_msg_error(_id,"Cannot save product %s", CPM_PROCATG);
            freeplist(qclist);
            freeplist(alist);
            freespace(outcpm);
            freespace(outmdome);
            omega_mdome_tidy();
            return -1;;
		}

        freeplist(qclist);
        freeplist(alist);
        freeimage(ps.cpixels);
        freefits(ps.firstdome);
        ps.firstdome = NULL;
        ps.eh = NULL;

    }    /* go and work on next extension */

    /*Clean up */

    freespace(outcpm);
    freespace(outmdome);
    omega_mdome_tidy();

    return 0;

}

/*----------------------------------------------------------------------------*/
/**
  @brief
    Trim and combine input images

  @param    pars     the input list of parameters
  @param    xn       extension number

  @return
    0 if everything is ok, 1 if this detector is dead and -1 otherwise
 */
/*----------------------------------------------------------------------------*/
int omega_mdome_combine(cpl_parameterlist *pars, int xn)
{

    int i,nflats,live, naxis1, naxis2;
    float bias = BIAS;
    double gain = 0.0;
    double median = 1.0;
    double threshold = 0.0;
    double *data_scales = NULL;
    const char *_id = "omega_mdome_combine";
    const char *mdome_name = "omega_mdome_temp.fits";
    const char *backname = "omega_mdome_background.fits";

    cpl_error_code code;
    const cpl_frame *domefr = NULL;
    cpl_vector *scales = NULL;
    cpl_vector *median_vector = NULL;
    cpl_image *mbias = NULL;
    cpl_image *trim_raw = NULL, *back_image = NULL, *dev = NULL;
    cpl_image *good_float = NULL, *good_int = NULL, *median_all = NULL;
    cpl_image *sum_data = NULL, *sum_good = NULL, *new_image = NULL;
    cpl_image *norm_image = NULL;
    cpl_imagelist *ilist = NULL;
    cpl_mask  	*good = NULL, *bpm_map = NULL;
    cpl_mask	*pixelmap = NULL;

    nflats = cpl_frameset_get_size(ps.domelist);
    domefr = cpl_frameset_get_first_const(ps.domelist);

    /* Check that this detector is live*/
    omega_pfits_get_detlive(ps.eh,&live);
    if (! live) {
        return 1;
    }

    /* Load master bias image */
    if(ps.mbframe != NULL){
        mbias = cpl_image_load(cpl_frame_get_filename(ps.mbframe), CPL_TYPE_FLOAT,0,xn);
        if (mbias == NULL) {
            cpl_msg_warning(_id,"Cannot load image %s", OMEGA_CALIB_BIAS);
        }
    }

     /*Create Mask from Hot and Cold Map */
    bpm_map = makebpm(ps.hframe, ps.cframe, xn);


    scales = cpl_vector_new (nflats);
    median_vector = cpl_vector_new (nflats);
    cpl_vector_fill(median_vector,0.0);

    data_scales = cpl_vector_get_data(scales);

    ilist = cpl_imagelist_new();
    cpl_msg_info (_id,"Doing trim and overscan correction on images");

    /* Loop through all images in frame list */
    for (i=0; i< nflats; i++){

        trim_raw = TrimOscanCorrect(domefr, omega_mdome_config.oc, xn);
        if(trim_raw == NULL){
            freevector(scales);
            freevector(median_vector);
            freeilist(ilist);
            return -1;
        }

        if (mbias != NULL)
            cpl_image_subtract(trim_raw, mbias);
        else
            cpl_image_subtract_scalar(trim_raw, bias);

        if (bpm_map == NULL) {
            ps.stats = cpl_stats_new_from_image(trim_raw, CPL_STATS_ALL);
            median = cpl_stats_get_median(ps.stats);
        }
        else {
            cpl_image_reject_from_mask(trim_raw, bpm_map);
            ps.stats = cpl_stats_new_from_image(trim_raw, CPL_STATS_ALL);
            median = cpl_stats_get_median(ps.stats);
        }

        cpl_vector_set(median_vector, i, median);

        data_scales[i] = (double)1.0/median;

        cpl_image_divide_scalar(trim_raw, median);

        cpl_imagelist_set(ilist, trim_raw, i);


        domefr = cpl_frameset_get_next_const(ps.domelist);
        if (domefr == NULL)
            break;

        freestats(ps.stats);

    }

    omega_mdome_config.RawMin    = cpl_vector_get_min(median_vector);
    omega_mdome_config.RawMax    = cpl_vector_get_max(median_vector);
    omega_mdome_config.RawMean   = cpl_vector_get_mean(median_vector);
    omega_mdome_config.RawMedian = cpl_vector_get_median(median_vector);
    omega_mdome_config.RawStdev  = cpl_vector_get_stdev(median_vector);

   freevector(median_vector);

    freestats(ps.stats);
    freeimage(mbias);
    freemask(bpm_map);

    if (ilist == NULL) {
        cpl_msg_error(_id,"Error in image list <%s>",cpl_error_get_message());
        freevector(scales);
        freeimage(trim_raw);
        return -1;
    }


    /*  Read gain from the header */
    /* FIXME: or read gain from gain recipe product?? */

    /* The images have the CONAD as the correct value for
     * the gain, therefore the following lines should be
     * changed if the headers are fixed.
     */

//    omega_pfits_get_gain(ps.eh, &gain);
    omega_pfits_get_conad(ps.eh, &gain);
    /*FIXME: should we avoid having gain=0?*/
    if (gain < 0.1){
    	omega_pfits_get_gain(ps.eh, &gain);
//    	omega_pfits_get_conad(ps.eh, &gain);
    	if(gain < 0.1) {
    		cpl_msg_info(_id,"Using default value of gain");
    		gain = 2.0;
    	}
    }
    else {
        cpl_msg_info(_id,"Gain value from image is %g", gain);
    }

    cpl_msg_info (_id,"Computing the median of all images...");
    median_all = cpl_imagelist_collapse_median_create(ilist);

    if (median_all == NULL) {
        cpl_msg_error (_id,"Cannot take median of list <%s>",cpl_error_get_message());
        freevector(scales);
        freeilist(ilist);
        return -1;
    }

    naxis1 = cpl_image_get_size_x(median_all);
    naxis2 = cpl_image_get_size_y(median_all);

    sum_data = cpl_image_new(naxis1,naxis2,CPL_TYPE_FLOAT);
    sum_good = cpl_image_new(naxis1,naxis2,CPL_TYPE_FLOAT);

    /* Clip the outliers from the images*/
    for (i=0; i< nflats; i++){

        trim_raw = cpl_imagelist_get(ilist, i);
        new_image = cpl_image_duplicate(trim_raw);

        /* FIXME: setting negative values to 0 to avoid failure of SQRT in next step
         * when there are negative values in the image. These negative values happen when any
         * given pixel in the domeflat is smaller than the same pixel in the master_bias image.
         */
        cpl_image_threshold(trim_raw, 0, FLT_MAX, 0, 2e20);

        dev = cpl_image_subtract_create(median_all, trim_raw);

        code = cpl_image_power(trim_raw, 0.5);
        if(code != CPL_ERROR_NONE) {
            cpl_msg_error(_id,"Error in SQRT operation <%s>",cpl_error_get_message());
            freevector(scales);
            freeilist(ilist);
            freeimage(median_all);
            freeimage(sum_data);
            freeimage(sum_good);
            freeimage(new_image);
            freeimage(dev);
            return -1;
        }

        cpl_image_divide(dev, trim_raw);

        threshold = omega_mdome_config.sigma * (sqrt(gain * data_scales[i]));

        good = cpl_mask_threshold_image_create(dev, -threshold, threshold);

        freeimage(dev);
        good_int = cpl_image_new_from_mask(good) ;
        freemask(good) ;

        good_float = cpl_image_cast(good_int, CPL_TYPE_FLOAT);
        freeimage(good_int);

        cpl_image_multiply(new_image, good_float);
        cpl_image_add(sum_data, new_image);

        cpl_image_add(sum_good, good_float);

        freeimage(new_image);
        freeimage(good_float);

    }

    freeimage(median_all);
    freevector(scales);

    /* Create master dome flat */
    ps.mdome = cpl_image_divide_create(sum_data, sum_good);
    if (ps.mdome == NULL) {
        cpl_msg_error(_id,"Error in division %s <%s>",MDOME_PROCATG, cpl_error_get_message());
        freeilist(ilist);
        freeimage(sum_data);
        freeimage(sum_good);
        return -1;
    }

    freeimage(sum_data);
    freeimage(sum_good);
    freeilist(ilist);

    /* Save temporary master dome to disk to use in Sextractor */
    cpl_image_save(ps.mdome, mdome_name, BITPIX, NULL, CPL_IO_DEFAULT);


    /* Create New Cold Pixels Map */
    /* Call function to create background image*/
    if(omega_create_background(pars, mdome_name, backname) != 0){
        cpl_msg_warning(_id,"Error in creating background image");
/*		return -1;*/
    }

    back_image = cpl_image_load(backname, CPL_TYPE_FLOAT, 0, 0);
    if (back_image != NULL){
        norm_image = cpl_image_divide_create(ps.mdome, back_image);
        freeimage(back_image);
    }
    else{
        cpl_msg_warning(_id,"Could not create normalised image. Cannot load background image.");
        norm_image = NULL;
    }

    /*Explanation for next steps: the image is normalized to around 1.0, therefore
    the bad pixels will be all pixels OUTSIDE the range lthre--hthre. Because of
    this, a mask_not needs to be applied to set the bad pixels to CPL_BINARY_1
    and count them using cpl_mask_count (which counts CPL_BINARY_1)*/

    /* This cpl function creates a mask of all the pixels within the given interval.
     * For example if an image has pixels in the interval 0.0 - 1.0 and the function
     * is called like this cpl_mask_threshold_image_create(img,0.5,1.0), all the pixels
     * in the range 0.5 to 1.0 will be set to CPL_BINARY_1. The other pixels,
     * < 0.5 and > 1.0 will be set to CPL_BINARY_0.
     */
    pixelmap = cpl_mask_threshold_image_create(norm_image, omega_mdome_config.lthre,
                                               omega_mdome_config.hthre);
    cpl_mask_not(pixelmap);

    cpl_msg_info(_id,"Creating pixel map with thresholds: %g,%g",omega_mdome_config.lthre,
                 omega_mdome_config.hthre);

    omega_mdome_config.CountColdPixels = cpl_mask_count(pixelmap);
    if (omega_mdome_config.CountColdPixels == -1) {
        cpl_msg_warning(_id,"Pixelmap is NULL. <%s>", cpl_error_get_message());
    }

    cpl_msg_info(_id,"Detected %d cold pixels", omega_mdome_config.CountColdPixels);

    freeimage(norm_image);

    /* Convert mask to image to save later */
    ps.cpixels = cpl_image_new_from_mask(pixelmap);
    freemask(pixelmap);

    /* Compute statistics on master dome flat */
    ps.stats = cpl_stats_new_from_image(ps.mdome, CPL_STATS_ALL);
    if(ps.stats == NULL && (cpl_error_get_code() == CPL_ERROR_DATA_NOT_FOUND)){
        cpl_msg_warning(_id," There are no good pixels in image for doing statistics");
        cpl_error_reset();
    }
    else{
        omega_mdome_config.Mean = cpl_stats_get_mean(ps.stats);
        omega_mdome_config.Median = cpl_stats_get_median(ps.stats);
        omega_mdome_config.Stdev = cpl_stats_get_stdev(ps.stats);
    }

    freestats(ps.stats);


    return 0;
}


/* Initialize the pointers */
static void omega_mdome_init(void) {
    ps.labels = NULL;
    ps.mbframe = NULL;
    ps.cframe = NULL;
    ps.hframe = NULL;
    ps.domelist = NULL;
    ps.ph = NULL;
    ps.eh = NULL;
    ps.stats = NULL;
    ps.mdome = NULL;
    ps.cpixels = NULL;
    ps.firstdome = NULL;
}

/* Free any allocated memory */
static void omega_mdome_tidy(void) {
    freespace(ps.labels);
    freeframeset(ps.domelist);
    freestats(ps.stats);
    freeimage(ps.mdome);
    freeimage(ps.cpixels);
    freefits(ps.firstdome);
}
/**@}*/
