/* $Id: omega_readnoise.c,v 1.3 2012-03-26 08:42:09 agabasch Exp $
 *
 * This file is part of the OMEGA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: agabasch $
 * $Date: 2012-03-26 08:42:09 $
 * $Revision: 1.3 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include "omega_recipe.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup omega_readnoise  Read Noise
 *
 * @par Synopsis:
 * Calculate the read noise of the detector. The read_noise is computed by subtracting two raw bias images, \n
 * and dividing the standard deviation by sqrt(2). The statistics of the difference between the raw biases is \n
 * computed iteratively. Bad pixels are rejected if they differ more than process_params.REJECTION_THRESHOLD \n
 * sigma from the median. The maximum number of iteration is process_params.MAXIMUM_ITERATIONS. The recipes \n
 * expects the following tags in the set of frames:
 *
 * @par Input frames:
 * BIAS_READNOISE\n
 *
 * @par Output frames:
 * READNOISE_ADU
 *
 * @code
 *      #include omega_recipe.h
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int omega_readnoise_create(cpl_plugin *) ;
static int omega_readnoise_exec(cpl_plugin *) ;
static int omega_readnoise_destroy(cpl_plugin *) ;
static int omega_readnoise(cpl_frameset *,cpl_parameterlist *) ;

/*-----------------------------------------------------------------------------
                            Private Functions
 -----------------------------------------------------------------------------*/
int omega_readnoise_save(cpl_frameset *set, cpl_parameterlist *parlist);
static void omega_readnoise_init(void);
static void omega_readnoise_tidy(void);

static struct {
    /* Inputs. Parameters */
    float       rej_sig;
    int         niter;
    int         extnum;
    int         paf;

    /* Outputs. QC parameters */
    double	rnoise;
    double	mean_diff;
    double	median_diff;
    double  RawBias1Min;
    double  RawBias1Max;
    double  RawBias1Mean;
    double  RawBias1Median;
    double  RawBias1Stdev;

    double  RawBias2Min;
    double  RawBias2Max;
    double  RawBias2Mean;
    double  RawBias2Median;
    double  RawBias2Stdev;

}omega_readnoise_config;

static struct {
    cpl_size         *labels;
    cpl_frameset     *biaslist;
    omega_fits        *biasfits1;
    omega_fits        *biasfits2;
    cpl_stats        *stats;

    /*Products*/
    cpl_table        *table;
} ps;

/* Static variables */

/* Static variables */
#define RECIPE "omega_readnoise"

static cpl_frame *product_frame = NULL;
static int isfirst;
static int dummy;
static char outfile[100];
static char outpaf[100];
static const char *PROCATG = OMEGA_CALIB_RDNOISE;


/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module.
  @param    list    the plugin list
  @return   0 if everything is ok

  This function is exported.
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist * list)
{
    cpl_recipe  *   recipe = cpl_calloc(1, sizeof(*recipe)) ;
    cpl_plugin  *   plugin = &recipe->interface ;

    cpl_plugin_init(plugin,
                    CPL_PLUGIN_API,
                    OMEGA_BINARY_VERSION,
                    CPL_PLUGIN_TYPE_RECIPE,
                    "omega_readnoise",
                    "OMEGA - Calculate the read noise of the detector.",
                    "The read_noise is computed by subtracting two raw bias images, \n"
                    "and dividing the standard deviation by sqrt(2). The descriptor \n"
                    "value read_noise will be set to the computed value. The statistics \n"
                    "of the difference between the raw biases is computed iteratively. Bad \n"
                    "pixels are rejected if they differ more than process_params.REJECTION_THRESHOLD \n"
                    "sigma from the median. The maximum number of iteration is \n"
                    "process_params.MAXIMUM_ITERATIONS.",
                   "Sandra Castro",
                    "scastro@eso.org",
                    omega_get_license(),
                    omega_readnoise_create,
                    omega_readnoise_exec,
                    omega_readnoise_destroy) ;

    cpl_pluginlist_append(list, plugin) ;

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options
  @param    plugin  the plugin
  @return   0 if everything is ok

  Create the recipe instance and make it available to the application using the
  interface.
 */
/*----------------------------------------------------------------------------*/
static int omega_readnoise_create(cpl_plugin * plugin)
{
    cpl_recipe * recipe;
    cpl_parameter * p ;

    /* Do not create the recipe if an error code is already set */
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());
        return (int)cpl_error_get_code();
    }

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    /* Create the parameters list in the cpl_recipe object */
    recipe->parameters = cpl_parameterlist_new() ;

    if (recipe->parameters == NULL) {
        cpl_msg_error(cpl_func, "Parameter list allocation failed");
        cpl_ensure_code(0, (int)CPL_ERROR_ILLEGAL_OUTPUT);
    }

    /* Fill the parameters list */
    p = cpl_parameter_new_value("omega.omega_readnoise.ExtensionNumber",
                                CPL_TYPE_INT,
                                "FITS extension number to load (1 to 32). (-1 = all)",
                                "omega_readnoise",
                                -1) ;

    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"ext") ;
    cpl_parameterlist_append(recipe->parameters, p) ;

    p = cpl_parameter_new_value("omega.omega_readnoise.PAF",
                                CPL_TYPE_BOOL,
                                "Boolean value to create PAF files. 1(Yes), 0(No)",
                                "omega_readnoise",
                                0) ;

    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "paf") ;
    cpl_parameterlist_append(recipe->parameters, p) ;

    p = cpl_parameter_new_range("omega.omega_readnoise.NumberIter",
                            CPL_TYPE_INT,
                            "Maximum number of iterations",
                            "omega_readnoise",
                            5, 2, 10) ;

    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI,"niter") ;
    cpl_parameterlist_append(recipe->parameters, p) ;


    p = cpl_parameter_new_range("omega.omega_readnoise.RejSigma",
                                CPL_TYPE_DOUBLE,
                                "Sigma Clipping Threshold",
                                "omega_readnoise",
                                5.0, 1.0, 10.0) ;

    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI, "sig-clip") ;
    cpl_parameterlist_append(recipe->parameters, p) ;


    /* Return */
    return 0;
}


/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int omega_readnoise_exec(cpl_plugin * plugin)
{
    cpl_recipe * recipe;
    int recipe_status;

    /* Return immediately if an error code is already set */
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());
        return (int)cpl_error_get_code();
    }

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    /* Verify parameter and frame lists */
    if (recipe->parameters == NULL) {
        cpl_msg_error(cpl_func, "Recipe invoked with NULL parameter list");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }
    if (recipe->frames == NULL) {
        cpl_msg_error(cpl_func, "Recipe invoked with NULL frame set");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Invoke the recipe */
    recipe_status = omega_readnoise(recipe->frames, recipe->parameters);

    /* Ensure DFS-compliance of the products */
    if (cpl_dfs_update_product_header(recipe->frames)) {
        if (!recipe_status) recipe_status = (int)cpl_error_get_code();
    }

    return recipe_status;

}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int omega_readnoise_destroy(cpl_plugin * plugin)
{
    cpl_recipe  *recipe;

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    cpl_parameterlist_delete(recipe->parameters);

    return 0 ;
}


/**********************************************************************/
/**
  @brief    Do a read_noise measurement
  @param    raw_bias_frames     		imset with two trimmed bias images.
  @param    REJECTION_THRESHOLD        for bad pixels (sigma)
  @param    MAXIMUM_ITERATIONS         for estimating the statistics

 */
/*--------------------------------------------------------------------------*/

static int omega_readnoise(cpl_frameset *set,cpl_parameterlist *pars)
{

    int j,jst,jfn,live;
    cpl_size nlab;
    float stdev = 0.0;

    const char *_id = "omega_readnoise";
    const char *chipid = NULL;

    const cpl_frame    *bias1, *bias2;
    const cpl_image    *rawbiasfits1, *rawbiasfits2=NULL;
    cpl_image 		   *biasim;
    cpl_parameter 	   *par;
    cpl_propertylist *plist;


 /*Start the recipe */

    if (pars == NULL) {
        cpl_msg_error (_id, "Parameters list not found");
        return -1;
    }

    if (cpl_frameset_is_empty(set) == 1) {
        cpl_msg_error (_id, "Frameset not found");
        return -1;
    }

/* Initialise a few things */

    omega_readnoise_init();


/* Retrieve input parameters */

    par = cpl_parameterlist_find(pars, "omega.omega_readnoise.ExtensionNumber") ;
    omega_readnoise_config.extnum = cpl_parameter_get_int(par) ;

    par = cpl_parameterlist_find(pars, "omega.omega_readnoise.NumberIter") ;
    omega_readnoise_config.niter = cpl_parameter_get_int(par) ;

    par = cpl_parameterlist_find(pars, "omega.omega_readnoise.RejSigma") ;
    omega_readnoise_config.rej_sig = cpl_parameter_get_double(par) ;

    par = cpl_parameterlist_find(pars, "omega.omega_readnoise.PAF") ;
    omega_readnoise_config.paf = cpl_parameter_get_bool(par) ;

/* Identify the RAW and CALIB frames in the input frameset */
    if (oc_dfs_set_groups(set)) {
        cpl_msg_error(_id, "Cannot identify RAW and CALIB frames") ;
        omega_readnoise_tidy();
        return -1 ;
    }

/* Verify the frameset contents. */
/*Make sure there are at least 2 bias */

    if ((ps.labels = cpl_frameset_labelise(set,omega_compare_tags,
					   &nlab)) == NULL) {
        cpl_msg_error(_id,"Cannot labelise the input frameset");
        omega_readnoise_tidy();
        return -1;
    }
    if ((ps.biaslist = omega_frameset_subgroup(set,ps.labels,nlab,
						RNOISE_RAW)) == NULL) {
        cpl_msg_error(_id,"Cannot find bias frames in input frameset");
        omega_readnoise_tidy();
        return -1;
    }
    if (cpl_frameset_get_size(ps.biaslist) < 2) {
	cpl_msg_error(_id,"Need exactly 2 (%s) frames to run this recipe",RNOISE_RAW);
        omega_readnoise_tidy();
        return -1;
    }

    /* Set up some convenience variables */
    bias1 = cpl_frameset_get_frame_const(ps.biaslist,0);
    bias2 = cpl_frameset_get_frame_const(ps.biaslist,1);

    cpl_msg_info (_id,"Using %s frames: %s and %s",RNOISE_RAW,cpl_frame_get_filename(bias1),
                                                    cpl_frame_get_filename(bias2));

/* Loop for each of the image extensions */
    omega_exten_range(omega_readnoise_config.extnum,&jst,&jfn);
    if(omega_readnoise_config.extnum == 0){
    	cpl_msg_error(cpl_func,"Unsupported extension request, %d",omega_readnoise_config.extnum);
    	omega_readnoise_tidy();
    	return -1;
    }

    /* Check which instrument */
    if(omega_pfits_check_instrume(cpl_frameset_get_first_const(set)) == 1 &&
                  omega_readnoise_config.extnum == 0 && jfn == 32)
        jfn = 8;

    for (j = jst; j <= jfn; j++) {
        cpl_msg_info(_id,"Beginning work on extension %d",j);
        isfirst = (j == jst);
        omega_readnoise_config.rnoise = 0.0;
        omega_readnoise_config.mean_diff = 0.0;
        omega_readnoise_config.median_diff = 0.0;
        omega_readnoise_config.RawBias1Min=0.;
        omega_readnoise_config.RawBias1Max=0.;
        omega_readnoise_config.RawBias1Mean=0.;
        omega_readnoise_config.RawBias1Median=0.;
        omega_readnoise_config.RawBias1Stdev=0.;
        omega_readnoise_config.RawBias2Min=0.;
        omega_readnoise_config.RawBias2Max=0.;
        omega_readnoise_config.RawBias2Mean=0.;
        omega_readnoise_config.RawBias2Median=0.;
        omega_readnoise_config.RawBias2Stdev=0.;

        ps.biasfits1 = omega_fits_load(bias1,CPL_TYPE_FLOAT,j);
        ps.biasfits2 = omega_fits_load(bias2,CPL_TYPE_FLOAT,j);

        if (ps.biasfits1 == NULL || ps.biasfits2 == NULL) {
            cpl_msg_error(_id,"NULL image input for extension %d",j);
            /*save dummy extension*/
            freefits(ps.biasfits1);
            freefits(ps.biasfits2);
            continue;
        }


        /* Check that this detector is live*/
    	plist = cpl_propertylist_load(cpl_frame_get_filename(bias1),j);
    	omega_pfits_get_detlive(plist,&live);
    	if (! live) {
    	    cpl_msg_warning(_id,"First bias image detector not live");
             /*save dummy extension*/
            freefits(ps.biasfits1);
            freefits(ps.biasfits2);
            freeplist(plist);
            continue;
         }
        freeplist(plist);

        plist = cpl_propertylist_load(cpl_frame_get_filename(bias2),j);
        omega_pfits_get_detlive(plist,&live);
    	if (! live) {
            cpl_msg_warning(_id,"Second bias image detector not live");
             /*save dummy extension*/
            freefits(ps.biasfits1);
            freefits(ps.biasfits2);
            freeplist(plist);
            continue;
        }
        freeplist(plist);

        /* Get the raw images and derive some QC parameter */

        rawbiasfits1=omega_fits_get_image(ps.biasfits1);
        rawbiasfits2=omega_fits_get_image(ps.biasfits2);
        omega_readnoise_config.RawBias1Min   =cpl_image_get_min(rawbiasfits1);
        omega_readnoise_config.RawBias1Max   =cpl_image_get_max(rawbiasfits1);
        omega_readnoise_config.RawBias1Mean  =cpl_image_get_mean(rawbiasfits1);
        omega_readnoise_config.RawBias1Median=cpl_image_get_median(rawbiasfits1);
        omega_readnoise_config.RawBias1Stdev =cpl_image_get_stdev(rawbiasfits1);
        omega_readnoise_config.RawBias2Min   =cpl_image_get_min(rawbiasfits2);
        omega_readnoise_config.RawBias2Max   =cpl_image_get_max(rawbiasfits2);
        omega_readnoise_config.RawBias2Mean  =cpl_image_get_mean(rawbiasfits2);
        omega_readnoise_config.RawBias2Median=cpl_image_get_median(rawbiasfits2);
        omega_readnoise_config.RawBias2Stdev =cpl_image_get_stdev(rawbiasfits2);


        /* Get the difference image */
        biasim = cpl_image_subtract_create(omega_fits_get_image(ps.biasfits1),
                                         omega_fits_get_image(ps.biasfits2));

        freefits(ps.biasfits2);

        /*Calculate statistics iteratively*/
        ps.stats = omega_iter_stat(biasim,omega_readnoise_config.rej_sig,omega_readnoise_config.niter);

        /* Now the read noise */
        if(ps.stats != NULL){
            stdev = cpl_stats_get_stdev(ps.stats);
            omega_readnoise_config.rnoise = stdev / sqrt(2);
            omega_readnoise_config.mean_diff = cpl_stats_get_mean(ps.stats);
            omega_readnoise_config.median_diff = cpl_stats_get_median(ps.stats);
        }
        else{
            cpl_msg_warning(_id,"Cannot calculate statistics iteratively");
        }

        plist = cpl_propertylist_load(cpl_frame_get_filename(bias1),j);
        chipid = omega_pfits_get_chipid(plist);

    	cpl_msg_info(_id,"The read noise of %s is: %5.3g", chipid, omega_readnoise_config.rnoise);
        cpl_msg_info(_id,"Difference between bias frames is: %5.3g (mean), %5.3g (median)",
                         omega_readnoise_config.mean_diff, omega_readnoise_config.median_diff);

        freeimage(biasim);
        freestats(ps.stats);
        freeplist(plist);

        /* Create the product */
        ps.table = cpl_table_new(1);
        cpl_table_new_column(ps.table, "READNOISE", CPL_TYPE_DOUBLE);
    	cpl_table_new_column(ps.table, "MEAN_DIFF", CPL_TYPE_DOUBLE);
    	cpl_table_new_column(ps.table, "MEDIAN_DIFF", CPL_TYPE_DOUBLE);
    	cpl_table_set_double(ps.table, "READNOISE", 0, omega_readnoise_config.rnoise);
    	cpl_table_set_double(ps.table, "MEAN_DIFF", 0, omega_readnoise_config.mean_diff);
    	cpl_table_set_double(ps.table, "MEDIAN_DIFF", 0, omega_readnoise_config.median_diff);

        /* Save the product */
        if(omega_readnoise_save(set, pars) == -1){
            cpl_msg_error(_id,"Cannot save this extension product");
            omega_readnoise_tidy();
            return -1;
        }

        freefits(ps.biasfits1);
        freetable(ps.table);
    }  /* Work on next extension */

    omega_readnoise_tidy();

    return 0;

}

/*----------------------------------------------------------------------------*/
/**
  @brief    Save the omega_readnoise recipe products on disk

  @param    set         the input frame set
  @param    parlist     the input list of parameters

  @return   0 if everything is ok, -1 otherwise
 */
/*----------------------------------------------------------------------------*/
int omega_readnoise_save(cpl_frameset *set,
                               cpl_parameterlist *parlist)
{

    const char *fctid = "omega_readnoise_save";
    cpl_propertylist *plist;
    cpl_propertylist *qclist;


    /* If it is the first time, setup initial frame */
    if (isfirst) {

        /* Get the name of the instrument */
        if(omega_pfits_check_instrume(cpl_frameset_get_first_const(set)) == 1)
            sprintf(INSTRUME,"wfi");

        /* Set the file name*/
        sprintf(outfile, "%s_%s.fits",INSTRUME,PROCATG) ;

        /* Create a new product frame object and define some tags */
        product_frame = cpl_frame_new();
        cpl_frame_set_filename(product_frame,outfile);
        cpl_frame_set_tag(product_frame,PROCATG);
        cpl_frame_set_type(product_frame,CPL_FRAME_TYPE_TABLE);
        cpl_frame_set_group(product_frame,CPL_FRAME_GROUP_PRODUCT);
        cpl_frame_set_level(product_frame,CPL_FRAME_LEVEL_FINAL);

        plist = cpl_propertylist_new();

        /* Add DataFlow keywords in primary header */
        if (cpl_dfs_setup_product_header(plist, product_frame, set, parlist,
                                        RECIPE,PIPEID,DICID,NULL) != CPL_ERROR_NONE) {

            cpl_msg_warning(fctid, "Problem in the main header of product DFS-compliance") ;
        }

        /* Once CPL recalculates these values, the following
         * line should be removed
         */
        cpl_propertylist_erase_regexp(plist,REM_PRIM_KEYS,0);

        /* Save the empty primary unit */
        if (cpl_propertylist_save(plist,outfile,CPL_IO_DEFAULT) != CPL_ERROR_NONE){
            cpl_msg_error(fctid,"Cannot save product PHU");
            cpl_frame_delete(product_frame);
            freeplist(plist);
            return -1;
        }

        freeplist(plist);
        cpl_frameset_insert(set,product_frame);
    }

    /* Setup the extension */
    plist = cpl_propertylist_duplicate(omega_fits_get_ehu(ps.biasfits1));
    qclist = cpl_propertylist_new();

    /* Add DataFlow keywords in extension header */
    if (cpl_dfs_setup_product_header(plist, product_frame, set, parlist,
        	                         RECIPE,PIPEID,DICID,NULL) != CPL_ERROR_NONE) {

        cpl_msg_warning(fctid, "Problem in the extension header of product DFS-compliance") ;
    }

    /*Remove undesired keywords from extension header*/
    cpl_propertylist_erase_regexp(plist, REM_EXT_KEYS, 0);


    /* Now add the QC data... */

    cpl_propertylist_update_float(qclist,"ESO QC READNOISE",
				  omega_readnoise_config.rnoise);
    cpl_propertylist_update_float(qclist,"ESO QC READNOISE MEAN DIFF",
				  omega_readnoise_config.mean_diff);
    cpl_propertylist_update_float(qclist,"ESO QC READNOISE MEDIAN DIFF",
				  omega_readnoise_config.median_diff);

    cpl_propertylist_update_float(qclist,"ESO QC RAW BIAS1 MIN"   , omega_readnoise_config.RawBias1Min   );
    cpl_propertylist_update_float(qclist,"ESO QC RAW BIAS1 MAX"   , omega_readnoise_config.RawBias1Max   );
    cpl_propertylist_update_float(qclist,"ESO QC RAW BIAS1 MEAN"  , omega_readnoise_config.RawBias1Mean  );
    cpl_propertylist_update_float(qclist,"ESO QC RAW BIAS1 MEDIAN", omega_readnoise_config.RawBias1Median);
    cpl_propertylist_update_float(qclist,"ESO QC RAW BIAS1 STDEV" , omega_readnoise_config.RawBias1Stdev );
    cpl_propertylist_update_float(qclist,"ESO QC RAW BIAS2 MIN"   , omega_readnoise_config.RawBias2Min   );
    cpl_propertylist_update_float(qclist,"ESO QC RAW BIAS2 MAX"   , omega_readnoise_config.RawBias2Max   );
    cpl_propertylist_update_float(qclist,"ESO QC RAW BIAS2 MEAN"  , omega_readnoise_config.RawBias2Mean  );
    cpl_propertylist_update_float(qclist,"ESO QC RAW BIAS2 MEDIAN", omega_readnoise_config.RawBias2Median);
    cpl_propertylist_update_float(qclist,"ESO QC RAW BIAS2 STDEV ", omega_readnoise_config.RawBias2Stdev );

    cpl_propertylist_set_comment(qclist,"ESO QC RAW BIAS1 MIN"   , "minimum value of first frame");
    cpl_propertylist_set_comment(qclist,"ESO QC RAW BIAS1 MAX"   , "maximum value of first frame");
    cpl_propertylist_set_comment(qclist,"ESO QC RAW BIAS1 MEAN"  , "mean value of first frame");
    cpl_propertylist_set_comment(qclist,"ESO QC RAW BIAS1 MEDIAN", "median value of first frame");
    cpl_propertylist_set_comment(qclist,"ESO QC RAW BIAS1 STDEV" , "standard deviation of first frame");
    cpl_propertylist_set_comment(qclist,"ESO QC RAW BIAS2 MIN"   , "minimum value of second frame");
    cpl_propertylist_set_comment(qclist,"ESO QC RAW BIAS2 MAX"   , "maximum value of second frame");
    cpl_propertylist_set_comment(qclist,"ESO QC RAW BIAS2 MEAN"  , "mean value of second frame");
    cpl_propertylist_set_comment(qclist,"ESO QC RAW BIAS2 MEDIAN", "median value of second frame");
    cpl_propertylist_set_comment(qclist,"ESO QC RAW BIAS2 STDEV ", "standard deviation of second frame");

    /* Append it to the extension header */
    cpl_propertylist_append(plist, qclist);

    /* Add comments to QC parameters in extension header*/
    cpl_propertylist_set_comment(plist,"ESO QC READNOISE",
				 "Calculated detector readnoise");
    cpl_propertylist_set_comment(plist,"ESO QC READNOISE MEAN DIFF",
				 "Calculated mean of difference");
    cpl_propertylist_set_comment(plist,"ESO QC READNOISE MEDIAN DIFF",
				 "Calculated median of difference");

     /* 'Save' the product */
    if (cpl_table_save(ps.table,NULL,plist,outfile,CPL_IO_EXTEND) != CPL_ERROR_NONE) {
        cpl_msg_error(fctid,"Cannot save product. %s", cpl_error_get_message());
        freeplist(plist);
        freeplist(qclist);
        return -1;
    }

    /* Save the paf file */
    if(omega_readnoise_config.paf){
        int xtnum = omega_fits_get_extnum(ps.biasfits1);
        const char pafcopy[] = "^(DATE-OBS|ARCFILE|ESO TPL ID|ESO DET WIN1 DIT1|MJD-OBS|EXTNAME|ESO PRO CATG)$";
        sprintf(outpaf, "%s_%s_%d.paf",INSTRUME,PROCATG,xtnum) ;
        cpl_frame *fframe = cpl_frameset_get_first(set);
        cpl_propertylist *mlist = cpl_propertylist_load_regexp(cpl_frame_get_filename(fframe),0,pafcopy, 0);
        cpl_propertylist_copy_property_regexp(qclist, mlist, pafcopy, 0);
        if(cpl_propertylist_copy_property_regexp(qclist, plist,
                                             pafcopy, 0) != CPL_ERROR_NONE){
            cpl_msg_warning(fctid,"Some mandatory keywords are missing in PAF file");
        }

        cpl_dfs_save_paf(INSTRUME, RECIPE, qclist, outpaf);
        freeplist(mlist);
    }

    freeplist(plist);
    freeplist(qclist);

    return 0;
}


/*---------------------------------------------------------------------------*/
/**
  @brief    Initialise the pointers in the memory structure
 */
/*---------------------------------------------------------------------------*/

static void omega_readnoise_init(void) {
    ps.labels = NULL;
    ps.biaslist = NULL;
    ps.biasfits1 = NULL;
    ps.biasfits2 = NULL;
    ps.stats = NULL;
    ps.table = NULL;
    return;
}

/*---------------------------------------------------------------------------*/
/**
  @brief    Free any allocated workspace in the memory structure
 */
/*---------------------------------------------------------------------------*/

static void omega_readnoise_tidy(void) {
    freespace(ps.labels);
    freeframeset(ps.biaslist);
    freefits(ps.biasfits1);
    freefits(ps.biasfits2);
    freetable(ps.table);
    freestats(ps.stats);
    return;
}
/**@}*/

